import numpy as np
import torch
import torch.nn.functional as F
from .position import PositionEmbeddingSine


class InputPadder:
    """Pads images such that dimensions are divisible by 8"""

    def __init__(self, dims, mode="sintel", padding_factor=8):
        self.ht, self.wd = dims[-2:]
        pad_ht = (
            ((self.ht // padding_factor) + 1) * padding_factor - self.ht
        ) % padding_factor
        pad_wd = (
            ((self.wd // padding_factor) + 1) * padding_factor - self.wd
        ) % padding_factor
        if mode == "sintel":
            self._pad = [
                pad_wd // 2,
                pad_wd - pad_wd // 2,
                pad_ht // 2,
                pad_ht - pad_ht // 2,
            ]
        else:
            self._pad = [pad_wd // 2, pad_wd - pad_wd // 2, 0, pad_ht]

    def pad(self, *inputs):
        return [F.pad(x, self._pad, mode="replicate") for x in inputs]

    def unpad(self, x):
        ht, wd = x.shape[-2:]
        c = [self._pad[2], ht - self._pad[3], self._pad[0], wd - self._pad[1]]
        return x[..., c[0] : c[1], c[2] : c[3]]


def bilinear_sampler(img, coords, mode="bilinear", mask=False, padding_mode="zeros"):
    """Wrapper for grid_sample, uses pixel coordinates"""
    if coords.size(-1) != 2:  # [B, 2, H, W] -> [B, H, W, 2]
        coords = coords.permute(0, 2, 3, 1)

    H, W = img.shape[-2:]
    # H = height if height is not None else img.shape[-2]
    # W = width if width is not None else img.shape[-1]

    xgrid, ygrid = coords.split([1, 1], dim=-1)

    # To handle H or W equals to 1 by explicitly defining height and width
    if H == 1:
        assert ygrid.abs().max() < 1e-8
        H = 10
    if W == 1:
        assert xgrid.abs().max() < 1e-8
        W = 10

    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(
        img, grid, mode=mode, padding_mode=padding_mode, align_corners=True
    )

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.squeeze(-1).to(dtype=coords.dtype)

    return img


def coords_grid(batch, ht, wd, normalize=False, dtype=None, device=None):
    if normalize:  # [-1, 1]
        coords = torch.meshgrid(
            2 * torch.arange(ht, dtype=dtype, device=device) / (ht - 1) - 1,
            2 * torch.arange(wd, dtype=dtype, device=device) / (wd - 1) - 1,
            indexing="ij",
        )
    else:
        coords = torch.meshgrid(
            torch.arange(ht, dtype=dtype, device=device),
            torch.arange(wd, dtype=dtype, device=device),
            indexing="ij",
        )
    coords = torch.stack(coords[::-1], dim=0)
    return coords[None].repeat(batch, 1, 1, 1)  # [B, 2, H, W]


def coords_grid_np(h, w):  # used for accumulating high speed sintel flow testdata
    coords = np.meshgrid(
        np.arange(h, dtype=np.float32), np.arange(w, dtype=np.float32), indexing="ij"
    )
    coords = np.stack(coords[::-1], axis=-1)  # [H, W, 2]

    return coords


def compute_out_of_boundary_mask(flow, downsample_factor=None):
    # flow: [B, 2, H, W]
    assert flow.dim() == 4 and flow.size(1) == 2
    b, _, h, w = flow.shape
    init_coords = coords_grid(b, h, w).to(flow.device)
    corres = init_coords + flow  # [B, 2, H, W]

    if downsample_factor is not None:
        assert w % downsample_factor == 0 and h % downsample_factor == 0
        # the actual max disp can predict is in the downsampled feature resolution, then upsample
        max_w = (w // downsample_factor - 1) * downsample_factor
        max_h = (h // downsample_factor - 1) * downsample_factor
        # print('max_w: %d, max_h: %d' % (max_w, max_h))
    else:
        max_w = w - 1
        max_h = h - 1

    valid_mask = (
        (corres[:, 0] >= 0)
        & (corres[:, 0] <= max_w)
        & (corres[:, 1] >= 0)
        & (corres[:, 1] <= max_h)
    )

    # in case very large flow
    flow_mask = (flow[:, 0].abs() <= max_w) & (flow[:, 1].abs() <= max_h)

    valid_mask = valid_mask & flow_mask

    return valid_mask  # [B, H, W]


def normalize_coords(grid):
    """Normalize coordinates of image scale to [-1, 1]
    Args:
        grid: [B, 2, H, W]
    """
    assert grid.size(1) == 2
    h, w = grid.size()[2:]
    grid[:, 0, :, :] = 2 * (grid[:, 0, :, :].clone() / (w - 1)) - 1  # x: [-1, 1]
    grid[:, 1, :, :] = 2 * (grid[:, 1, :, :].clone() / (h - 1)) - 1  # y: [-1, 1]
    # grid = grid.permute((0, 2, 3, 1))  # [B, H, W, 2]
    return grid


def flow_warp(feature, flow, mask=False, padding_mode="zeros"):
    b, c, h, w = feature.size()
    assert flow.size(1) == 2

    grid = coords_grid(b, h, w).to(flow.device) + flow  # [B, 2, H, W]

    return bilinear_sampler(feature, grid, mask=mask, padding_mode=padding_mode)


def upflow8(flow, mode="bilinear"):
    new_size = (8 * flow.shape[2], 8 * flow.shape[3])
    return 8 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)


def bilinear_upflow(flow, scale_factor=8):
    assert flow.size(1) == 2
    flow = (
        F.interpolate(
            flow, scale_factor=scale_factor, mode="bilinear", align_corners=True
        )
        * scale_factor
    )

    return flow


def upsample_flow(flow, img):
    if flow.size(-1) != img.size(-1):
        scale_factor = img.size(-1) / flow.size(-1)
        flow = (
            F.interpolate(
                flow, size=img.size()[-2:], mode="bilinear", align_corners=True
            )
            * scale_factor
        )
    return flow


def count_parameters(model):
    num = sum(p.numel() for p in model.parameters() if p.requires_grad)
    return num


def set_bn_eval(m):
    classname = m.__class__.__name__
    if classname.find("BatchNorm") != -1:
        m.eval()


def generate_window_grid(
    h_min, h_max, w_min, w_max, len_h, len_w, dtype=None, device=None
):
    assert device is not None

    x, y = torch.meshgrid(
        [
            torch.linspace(w_min, w_max, len_w, dtype=dtype, device=device),
            torch.linspace(h_min, h_max, len_h, dtype=dtype, device=device),
        ],
        indexing="ij",
    )
    grid = torch.stack((x, y), -1).transpose(0, 1)  # [H, W, 2]

    return grid


def normalize_coords(coords, h, w):
    # coords: [B, H, W, 2]
    c = torch.Tensor([(w - 1) / 2.0, (h - 1) / 2.0]).to(
        dtype=coords.dtype, device=coords.device
    )
    return (coords - c) / c  # [-1, 1]


def normalize_img(img0, img1):
    # loaded images are in [0, 255]
    # normalize by ImageNet mean and std
    mean = (
        torch.tensor([0.485, 0.456, 0.406])
        .view(1, 3, 1, 1)
        .to(dtype=img1.dtype, device=img1.device)
    )
    std = (
        torch.tensor([0.229, 0.224, 0.225])
        .view(1, 3, 1, 1)
        .to(dtype=img1.dtype, device=img1.device)
    )
    img0 = (img0 - mean) / std
    img1 = (img1 - mean) / std

    return img0, img1


def split_feature(
    feature,
    num_splits=2,
    channel_last=False,
):
    if channel_last:  # [B, H, W, C]
        b, h, w, c = feature.size()
        assert h % num_splits == 0 and w % num_splits == 0

        b_new = b * num_splits * num_splits
        h_new = h // num_splits
        w_new = w // num_splits

        feature = (
            feature.view(b, num_splits, h // num_splits, num_splits, w // num_splits, c)
            .permute(0, 1, 3, 2, 4, 5)
            .reshape(b_new, h_new, w_new, c)
        )  # [B*K*K, H/K, W/K, C]
    else:  # [B, C, H, W]
        b, c, h, w = feature.size()
        assert h % num_splits == 0 and w % num_splits == 0

        b_new = b * num_splits * num_splits
        h_new = h // num_splits
        w_new = w // num_splits

        feature = (
            feature.view(b, c, num_splits, h // num_splits, num_splits, w // num_splits)
            .permute(0, 2, 4, 1, 3, 5)
            .reshape(b_new, c, h_new, w_new)
        )  # [B*K*K, C, H/K, W/K]

    return feature


def merge_splits(
    splits,
    num_splits=2,
    channel_last=False,
):
    if channel_last:  # [B*K*K, H/K, W/K, C]
        b, h, w, c = splits.size()
        new_b = b // num_splits // num_splits

        splits = splits.view(new_b, num_splits, num_splits, h, w, c)
        merge = (
            splits.permute(0, 1, 3, 2, 4, 5)
            .contiguous()
            .view(new_b, num_splits * h, num_splits * w, c)
        )  # [B, H, W, C]
    else:  # [B*K*K, C, H/K, W/K]
        b, c, h, w = splits.size()
        new_b = b // num_splits // num_splits

        splits = splits.view(new_b, num_splits, num_splits, c, h, w)
        merge = (
            splits.permute(0, 3, 1, 4, 2, 5)
            .contiguous()
            .view(new_b, c, num_splits * h, num_splits * w)
        )  # [B, C, H, W]

    return merge


def generate_shift_window_attn_mask(
    input_resolution,
    window_size_h,
    window_size_w,
    shift_size_h,
    shift_size_w,
    device=torch.device("cuda"),
):
    # ref: https://github.com/microsoft/Swin-Transformer/blob/main/models/swin_transformer.py
    # calculate attention mask for SW-MSA
    h, w = input_resolution
    img_mask = torch.zeros((1, h, w, 1)).to(device)  # 1 H W 1
    h_slices = (
        slice(0, -window_size_h),
        slice(-window_size_h, -shift_size_h),
        slice(-shift_size_h, None),
    )
    w_slices = (
        slice(0, -window_size_w),
        slice(-window_size_w, -shift_size_w),
        slice(-shift_size_w, None),
    )
    cnt = 0
    for h in h_slices:
        for w in w_slices:
            img_mask[:, h, w, :] = cnt
            cnt += 1

    mask_windows = split_feature(
        img_mask, num_splits=input_resolution[-1] // window_size_w, channel_last=True
    )

    mask_windows = mask_windows.view(-1, window_size_h * window_size_w)
    attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
    attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(
        attn_mask == 0, float(0.0)
    )

    return attn_mask


def feature_add_position(feature0, feature1, attn_splits, feature_channels):
    pos_enc = PositionEmbeddingSine(num_pos_feats=feature_channels // 2)

    if attn_splits > 1:  # add position in splited window
        feature0_splits = split_feature(feature0, num_splits=attn_splits)
        feature1_splits = split_feature(feature1, num_splits=attn_splits)

        position = pos_enc(feature0_splits)

        feature0_splits = feature0_splits + position
        feature1_splits = feature1_splits + position

        feature0 = merge_splits(feature0_splits, num_splits=attn_splits)
        feature1 = merge_splits(feature1_splits, num_splits=attn_splits)
    else:
        position = pos_enc(feature0)

        feature0 = feature0 + position
        feature1 = feature1 + position

    return feature0, feature1


def upsample_flow_with_mask(flow, up_mask, upsample_factor):
    # convex upsampling following raft

    mask = up_mask
    b, flow_channel, h, w = flow.shape
    mask = mask.view(
        b, 1, 9, upsample_factor, upsample_factor, h, w
    )  # [B, 1, 9, K, K, H, W]
    mask = torch.softmax(mask, dim=2)

    multiplier = upsample_factor
    up_flow = F.unfold(multiplier * flow, [3, 3], padding=1)
    up_flow = up_flow.view(b, flow_channel, 9, 1, 1, h, w)  # [B, 2, 9, 1, 1, H, W]

    up_flow = torch.sum(mask * up_flow, dim=2)  # [B, 2, K, K, H, W]
    up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)  # [B, 2, K, H, K, W]
    up_flow = up_flow.reshape(
        b, flow_channel, upsample_factor * h, upsample_factor * w
    )  # [B, 2, K*H, K*W]

    return up_flow


def split_feature_1d(
    feature,
    num_splits=2,
):
    # feature: [B, W, C]
    b, w, c = feature.size()
    assert w % num_splits == 0

    b_new = b * num_splits
    w_new = w // num_splits

    feature = feature.view(b, num_splits, w // num_splits, c).view(
        b_new, w_new, c
    )  # [B*K, W/K, C]

    return feature


def merge_splits_1d(
    splits,
    h,
    num_splits=2,
):
    b, w, c = splits.size()
    new_b = b // num_splits // h

    splits = splits.view(new_b, h, num_splits, w, c)
    merge = splits.view(new_b, h, num_splits * w, c)  # [B, H, W, C]

    return merge


def window_partition_1d(x, window_size_w):
    """
    Args:
        x: (B, W, C)
        window_size (int): window size
    Returns:
        windows: (num_windows*B, window_size, C)
    """
    B, W, C = x.shape
    x = x.view(B, W // window_size_w, window_size_w, C).view(-1, window_size_w, C)
    return x


def generate_shift_window_attn_mask_1d(
    input_w, window_size_w, shift_size_w, device=torch.device("cuda")
):
    # calculate attention mask for SW-MSA
    img_mask = torch.zeros((1, input_w, 1)).to(device)  # 1 W 1
    w_slices = (
        slice(0, -window_size_w),
        slice(-window_size_w, -shift_size_w),
        slice(-shift_size_w, None),
    )
    cnt = 0
    for w in w_slices:
        img_mask[:, w, :] = cnt
        cnt += 1

    mask_windows = window_partition_1d(img_mask, window_size_w)  # nW, window_size, 1
    mask_windows = mask_windows.view(-1, window_size_w)
    attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(
        2
    )  # nW, window_size, window_size
    attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(
        attn_mask == 0, float(0.0)
    )

    return attn_mask



import torch
from torch.autograd import Function
from torch.autograd import Variable

from ..build.lib import GANet


# import GANet
class NlfDownFunction(Function):
    @staticmethod
    def forward(ctx, input, g0):
        assert input.is_contiguous() == True and g0.is_contiguous() == True
        with torch.cuda.device_of(input):
            output_down = input.new().resize_(input.size()).zero_()
            GANet.nlf_down_cuda_forward(input, g0, output_down)
            output_down = output_down.contiguous()
        ctx.save_for_backward(input, g0, output_down)
        return output_down

    @staticmethod
    def backward(ctx, gradOutput):
        input, g0, output_down = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            gradInput = gradOutput.new().resize_(input.size()).zero_()
            grad0 = gradOutput.new().resize_(g0.size()).zero_()
            GANet.nlf_down_cuda_backward(
                input, g0, output_down, gradOutput, gradInput, grad0
            )
            gradInput = gradInput.contiguous()
            grad0 = grad0.contiguous()
        return gradInput, grad0


class NlfUpFunction(Function):
    @staticmethod
    def forward(ctx, input, g1):
        assert input.is_contiguous() == True and g1.is_contiguous() == True
        with torch.cuda.device_of(input):
            output_up = input.new().resize_(input.size()).zero_()
            GANet.nlf_up_cuda_forward(input, g1, output_up)
            output_up = output_up.contiguous()
        ctx.save_for_backward(input, g1, output_up)
        return output_up

    @staticmethod
    def backward(ctx, gradOutput):
        input, g1, output_up = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            gradInput = gradOutput.new().resize_(input.size()).zero_()
            grad1 = gradOutput.new().resize_(g1.size()).zero_()
            GANet.nlf_up_cuda_backward(
                input, g1, output_up, gradOutput, gradInput, grad1
            )
            gradInput = gradInput.contiguous()
            grad1 = grad1.contiguous()
        return gradInput, grad1


class NlfRightFunction(Function):
    @staticmethod
    def forward(ctx, input, g2):
        assert input.is_contiguous() == True and g2.is_contiguous() == True
        with torch.cuda.device_of(input):
            #            num, channels, height, width = input.size()
            #            output_right = input.new().resize_(num, channels, height, width).zero_()
            output_right = input.new().resize_(input.size()).zero_()
            GANet.nlf_right_cuda_forward(input, g2, output_right)
            output_right = output_right.contiguous()
        ctx.save_for_backward(input, g2, output_right)
        return output_right

    @staticmethod
    def backward(ctx, gradOutput):
        input, g2, output_right = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            #            num, channels, height, width = input.size()
            #            _, fsize, _, _ = g2.size()
            #            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            #            grad2 = gradOutput.new().resize_(num, fsize, height, width).zero_()
            gradInput = gradOutput.new().resize_(input.size()).zero_()
            grad2 = gradOutput.new().resize_(g2.size()).zero_()
            GANet.nlf_right_cuda_backward(
                input, g2, output_right, gradOutput, gradInput, grad2
            )
            gradInput = gradInput.contiguous()
            grad2 = grad2.contiguous()
        return gradInput, grad2


class NlfLeftFunction(Function):
    @staticmethod
    def forward(ctx, input, g3):
        assert input.is_contiguous() == True and g3.is_contiguous() == True
        with torch.cuda.device_of(input):
            #            num, channels, height, width = input.size()
            #            output_left = input.new().resize_(num, channels, height, width).zero_()
            output_left = input.new().resize_(input.size()).zero_()
            GANet.nlf_left_cuda_forward(input, g3, output_left)
            output_left = output_left.contiguous()
        ctx.save_for_backward(input, g3, output_left)
        return output_left

    @staticmethod
    def backward(ctx, gradOutput):
        input, g3, output_left = ctx.saved_tensors
        gradOutput = gradOutput.contiguous()
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            #            num, channels, height, width = input.size()
            #            _, fsize, _, _ = g3.size()
            #            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            #            grad3 = gradOutput.new().resize_(num, fsize, height, width).zero_()
            gradInput = gradOutput.new().resize_(input.size()).zero_()
            grad3 = gradOutput.new().resize_(g3.size()).zero_()
            GANet.nlf_left_cuda_backward(
                input, g3, output_left, gradOutput, gradInput, grad3
            )
            gradInput = gradInput.contiguous()
            grad3 = grad3.contiguous()
        return gradInput, grad3


class NlfFunction(Function):
    @staticmethod
    def forward(ctx, input, g0, g1, g2, g3):
        assert (
            input.is_contiguous() == True
            and g0.is_contiguous() == True
            and g1.is_contiguous() == True
            and g2.is_contiguous() == True
            and g3.is_contiguous() == True
        )
        with torch.cuda.device_of(input):
            num, channels, height, width = input.size()
            output_down = input.new().resize_(num, channels, height, width).zero_()
            output_up = input.new().resize_(num, channels, height, width).zero_()
            output_right = input.new().resize_(num, channels, height, width).zero_()
            output_left = input.new().resize_(num, channels, height, width).zero_()
            GANet.nlf_cuda_forward(
                input, g0, g1, g2, g3, output_down, output_up, output_right, output_left
            )
            #           GANet.sga_cuda_forward(input, filters, output, radius)

            output_down = output_down.contiguous()
            output_up = output_up.contiguous()
            output_right = output_right.contiguous()
            output_left = output_left.contiguous()
        ctx.save_for_backward(
            input, g0, g1, g2, g3, output_down, output_up, output_right, output_left
        )
        #        print(output_left.size())
        return output_left

    @staticmethod
    def backward(ctx, gradOutput):
        (
            input,
            g0,
            g1,
            g2,
            g3,
            output_down,
            output_up,
            output_right,
            output_left,
        ) = ctx.saved_tensors
        #        print temp_out.size()
        #        print mask.size()
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            num, channels, height, width = input.size()
            _, _, fsize, _, _ = g0.size()
            #            print fsize
            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            grad0 = (
                gradOutput.new().resize_(num, channels, fsize, height, width).zero_()
            )
            grad1 = (
                gradOutput.new().resize_(num, channels, fsize, height, width).zero_()
            )
            grad2 = (
                gradOutput.new().resize_(num, channels, fsize, height, width).zero_()
            )
            grad3 = (
                gradOutput.new().resize_(num, channels, fsize, height, width).zero_()
            )

            GANet.nlf_cuda_backward(
                input,
                g0,
                g1,
                g2,
                g3,
                output_down,
                output_up,
                output_right,
                output_left,
                gradOutput,
                gradInput,
                grad0,
                grad1,
                grad2,
                grad3,
            )
            #            GANet.lga_cuda_backward(input, filters, gradOutput, gradInput, gradFilters, radius)
            gradInput = gradInput.contiguous()
            grad0 = grad0.contiguous()
            grad1 = grad1.contiguous()
            grad2 = grad2.contiguous()
            grad3 = grad3.contiguous()
        return gradInput, grad0, grad1, grad2, grad3


class SgaFunction(Function):
    @staticmethod
    def forward(ctx, input, g0, g1, g2, g3):
        assert (
            input.is_contiguous() == True
            and g0.is_contiguous() == True
            and g1.is_contiguous() == True
            and g2.is_contiguous() == True
            and g3.is_contiguous() == True
        )
        with torch.cuda.device_of(input):
            num, channels, depth, height, width = input.size()
            output = input.new().resize_(num, channels, depth, height, width).zero_()
            temp_out = input.new().resize_(num, channels, depth, height, width).zero_()
            mask = input.new().resize_(num, channels, depth, height, width).zero_()
            GANet.sga_cuda_forward(input, g0, g1, g2, g3, temp_out, output, mask)
            #           GANet.sga_cuda_forward(input, filters, output, radius)

            output = output.contiguous()
        ctx.save_for_backward(input, g0, g1, g2, g3, temp_out, mask)
        return output

    @staticmethod
    def backward(ctx, gradOutput):
        input, g0, g1, g2, g3, temp_out, mask = ctx.saved_tensors
        #        print temp_out.size()
        #        print mask.size()
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            num, channels, depth, height, width = input.size()
            #            _, _, fsize, _, _ = g0.size()
            #            print fsize
            gradInput = (
                gradOutput.new().resize_(num, channels, depth, height, width).zero_()
            )
            grad0 = gradOutput.new().resize_(g0.size()).zero_()
            grad1 = gradOutput.new().resize_(g1.size()).zero_()
            grad2 = gradOutput.new().resize_(g2.size()).zero_()
            grad3 = gradOutput.new().resize_(g3.size()).zero_()
            temp_grad = (
                gradOutput.new().resize_(num, channels, depth, height, width).zero_()
            )
            max_idx = gradOutput.new().resize_(num, channels, height, width).zero_()

            GANet.sga_cuda_backward(
                input,
                g0,
                g1,
                g2,
                g3,
                temp_out,
                mask,
                max_idx,
                gradOutput,
                temp_grad,
                gradInput,
                grad0,
                grad1,
                grad2,
                grad3,
            )
            #            GANet.lga_cuda_backward(input, filters, gradOutput, gradInput, gradFilters, radius)
            gradInput = gradInput.contiguous()
            grad0 = grad0.contiguous()
            grad1 = grad1.contiguous()
            grad2 = grad2.contiguous()
            grad3 = grad3.contiguous()
        return gradInput, grad0, grad1, grad2, grad3


class Lga3d3Function(Function):
    @staticmethod
    def forward(ctx, input, filters, radius=1):
        ctx.radius = radius
        assert input.is_contiguous() == True and filters.is_contiguous() == True
        with torch.cuda.device_of(input):
            num, channels, depth, height, width = input.size()
            temp_out1 = input.new().resize_(num, channels, depth, height, width).zero_()
            temp_out2 = input.new().resize_(num, channels, depth, height, width).zero_()
            output = input.new().resize_(num, channels, depth, height, width).zero_()
            GANet.lga3d_cuda_forward(input, filters, temp_out1, radius)
            GANet.lga3d_cuda_forward(temp_out1, filters, temp_out2, radius)
            GANet.lga3d_cuda_forward(temp_out2, filters, output, radius)
            output = output.contiguous()
        ctx.save_for_backward(input, filters, temp_out1, temp_out2)
        return output

    @staticmethod
    def backward(ctx, gradOutput):
        input, filters, temp_out1, temp_out2 = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            num, channels, depth, height, width = input.size()
            _, _, fsize, _, _ = filters.size()
            #            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            gradFilters = (
                gradOutput.new().resize_(num, channels, fsize, height, width).zero_()
            )
            GANet.lga3d_cuda_backward(
                temp_out2, filters, gradOutput, temp_out2, gradFilters, ctx.radius
            )
            GANet.lga3d_cuda_backward(
                temp_out1, filters, temp_out2, temp_out1, gradFilters, ctx.radius
            )
            #            temp_out[...] = 0
            GANet.lga3d_cuda_backward(
                input, filters, temp_out1, temp_out2, gradFilters, ctx.radius
            )
            #            temp_out[...] = gradOutput[...]
            temp_out2 = temp_out2.contiguous()
            gradFilters = gradFilters.contiguous()
        return temp_out2, gradFilters, None


class Lga3d2Function(Function):
    @staticmethod
    def forward(ctx, input, filters, radius=1):
        ctx.radius = radius
        assert input.is_contiguous() == True and filters.is_contiguous() == True
        with torch.cuda.device_of(input):
            num, channels, depth, height, width = input.size()
            temp_out = input.new().resize_(num, channels, depth, height, width).zero_()
            output = input.new().resize_(num, channels, depth, height, width).zero_()
            GANet.lga3d_cuda_forward(input, filters, temp_out, radius)
            GANet.lga3d_cuda_forward(temp_out, filters, output, radius)
            output = output.contiguous()
        ctx.save_for_backward(input, filters, temp_out)
        return output

    @staticmethod
    def backward(ctx, gradOutput):
        input, filters, temp_out = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            num, channels, depth, height, width = input.size()
            _, _, fsize, _, _ = filters.size()
            #            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            gradFilters = (
                gradOutput.new().resize_(num, channels, fsize, height, width).zero_()
            )
            GANet.lga3d_cuda_backward(
                temp_out, filters, gradOutput, temp_out, gradFilters, ctx.radius
            )
            #            temp_out[...] = 0
            GANet.lga3d_cuda_backward(
                input, filters, temp_out, gradOutput, gradFilters, ctx.radius
            )
            temp_out[...] = gradOutput[...]
            temp_out = temp_out.contiguous()
            gradFilters = gradFilters.contiguous()
        return temp_out, gradFilters, None


class Lga3dFunction(Function):
    @staticmethod
    def forward(ctx, input, filters, radius=1):
        ctx.radius = radius
        ctx.save_for_backward(input, filters)
        assert input.is_contiguous() == True and filters.is_contiguous() == True
        with torch.cuda.device_of(input):
            num, channels, depth, height, width = input.size()
            output = input.new().resize_(num, channels, depth, height, width).zero_()
            GANet.lga3d_cuda_forward(input, filters, output, radius)
            output = output.contiguous()
        return output

    @staticmethod
    def backward(ctx, gradOutput):
        input, filters = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            num, channels, depth, height, width = input.size()
            _, _, fsize, _, _ = filters.size()
            gradInput = (
                gradOutput.new().resize_(num, channels, depth, height, width).zero_()
            )
            gradFilters = (
                gradOutput.new().resize_(num, channels, fsize, height, width).zero_()
            )
            GANet.lga3d_cuda_backward(
                input, filters, gradOutput, gradInput, gradFilters, ctx.radius
            )
            gradInput = gradInput.contiguous()
            gradFilters = gradFilters.contiguous()
        return gradInput, gradFilters, None


class Lga3Function(Function):
    @staticmethod
    def forward(ctx, input, filters, radius=1):
        ctx.radius = radius
        assert input.is_contiguous() == True and filters.is_contiguous() == True
        with torch.cuda.device_of(input):
            num, channels, height, width = input.size()
            temp_out1 = input.new().resize_(num, channels, height, width).zero_()
            temp_out2 = input.new().resize_(num, channels, height, width).zero_()
            output = input.new().resize_(num, channels, height, width).zero_()
            GANet.lga_cuda_forward(input, filters, temp_out1, radius)
            GANet.lga_cuda_forward(temp_out1, filters, temp_out2, radius)
            GANet.lga_cuda_forward(temp_out2, filters, output, radius)
            output = output.contiguous()
        ctx.save_for_backward(input, fitlers, temp_out1, temp_out2)
        return output

    @staticmethod
    def backward(ctx, gradOutput):
        input, filters, temp_out1, temp_out2 = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            num, channels, height, width = input.size()
            _, fsize, _, _ = filters.size()
            #            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            gradFilters = gradOutput.new().resize_(num, fsize, height, width).zero_()
            GANet.lga_cuda_backward(
                temp_out2, filters, gradOutput, temp_out2, gradFilters, ctx.radius
            )
            GANet.lga_cuda_backward(
                temp_out1, filters, temp_out2, temp_out1, gradFilters, ctx.radius
            )
            #            temp_out[...] = 0
            GANet.lga_cuda_backward(
                input, filters, temp_out1, temp_out2, gradFilters, ctx.radius
            )
            #            temp_out[...] = gradOutput[...]
            temp_out2 = temp_out2.contiguous()
            gradFilters = gradFilters.contiguous()
        return temp_out2, gradFilters, None


class Lga2Function(Function):
    @staticmethod
    def forward(ctx, input, filters, radius=1):
        ctx.radius = radius
        assert input.is_contiguous() == True and filters.is_contiguous() == True
        with torch.cuda.device_of(input):
            num, channels, height, width = input.size()
            temp_out = input.new().resize_(num, channels, height, width).zero_()
            output = input.new().resize_(num, channels, height, width).zero_()
            GANet.lga_cuda_forward(input, filters, temp_out, radius)
            GANet.lga_cuda_forward(temp_out, filters, output, radius)
            output = output.contiguous()
        ctx.save_for_backward(input, filters, temp_out)
        return output

    @staticmethod
    def backward(ctx, gradOutput):
        input, filters, temp_out = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            num, channels, height, width = input.size()
            _, fsize, _, _ = filters.size()
            #            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            gradFilters = gradOutput.new().resize_(num, fsize, height, width).zero_()
            GANet.lga_cuda_backward(
                temp_out, filters, gradOutput, temp_out, gradFilters, ctx.radius
            )
            #            temp_out[...] = 0
            GANet.lga_cuda_backward(
                input, filters, temp_out, gradOutput, gradFilters, ctx.radius
            )
            temp_out[...] = gradOutput[...]
            temp_out = temp_out.contiguous()
            gradFilters = gradFilters.contiguous()
        return temp_out, gradFilters, None


class Lgf2Function(Function):
    @staticmethod
    def forward(ctx, input, filters, radius=2):
        ctx.radius = radius
        assert input.is_contiguous() == True and filters.is_contiguous() == True
        with torch.cuda.device_of(input):
            #            num, channels, depth, height, width = input.size()
            #            temp_out = input.new().resize_(num, channels, depth, height, width).zero_()
            #            output = input.new().resize_(num, channels, depth, height, width).zero_()
            temp_out = input.new().resize_(input.size()).zero_()
            output = input.new().resize_(input.size()).zero_()
            GANet.lgf_cuda_forward(input, filters, temp_out, radius)
            GANet.lgf_cuda_forward(temp_out, filters, output, radius)
            output = output.contiguous()
        ctx.save_for_backward(input, filters, temp_out)
        return output

    @staticmethod
    def backward(ctx, gradOutput):
        input, filters, temp_out = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            #            num, channels, depth, height, width = input.size()
            #            _, fsize, _, _ = filters.size()
            #            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            #            gradFilters = gradOutput.new().resize_(num, fsize, height, width).zero_()
            gradFilters = gradOutput.new().resize_(filters.size()).zero_()
            GANet.lgf_cuda_backward(
                temp_out, filters, gradOutput, temp_out, gradFilters, ctx.radius
            )
            #            temp_out[...] = 0
            GANet.lgf_cuda_backward(
                input, filters, temp_out, gradOutput, gradFilters, ctx.radius
            )
            temp_out[...] = gradOutput[...]
            temp_out = temp_out.contiguous()
            gradFilters = gradFilters.contiguous()
        return temp_out, gradFilters, None


class LgaFunction(Function):
    @staticmethod
    def forward(ctx, input, filters):
        ctx.radius = radius
        assert input.is_contiguous() == True and filters.is_contiguous() == True
        with torch.cuda.device_of(input):
            num, channels, height, width = input.size()
            output = input.new().resize_(num, channels, height, width).zero_()
            GANet.lga_cuda_forward(input, filters, output, radius)
            output = output.contiguous()
        ctx.save_for_backward(input, filters)
        return output

    @staticmethod
    def backward(ctx, gradOutput):
        input, filters = ctx.saved_tensors
        assert gradOutput.is_contiguous() == True
        with torch.cuda.device_of(gradOutput):
            num, channels, height, width = input.size()
            _, fsize, _, _ = filters.size()
            gradInput = gradOutput.new().resize_(num, channels, height, width).zero_()
            gradFilters = gradOutput.new().resize_(num, fsize, height, width).zero_()
            GANet.lga_cuda_backward(
                input, filters, gradOutput, gradInput, gradFilters, ctx.radius
            )
            gradInput = gradInput.contiguous()
            gradFilters = gradFilters.contiguous()
        return gradInput, gradFilters, None


class MyLoss2Function(Function):
    @staticmethod
    def forward(ctx, input1, input2, thresh=1, alpha=2):
        ctx.thresh = thresh
        ctx.alpha = alpha
        diff = input1 - input2
        temp = torch.abs(diff)
        temp[temp < thresh] = temp[temp < thresh] ** 2 / thresh
        tag = (temp <= thresh + alpha) & (temp >= thresh)
        temp[tag] = temp[tag] * 2 - (temp[tag] - thresh) ** 2 / (2.0 * alpha) - thresh
        temp[temp > thresh + alpha] += alpha / 2.0
        ctx.save_for_backward(diff)
        return torch.mean(temp)

    @staticmethod
    def backward(ctx, gradOutput):
        (diff,) = ctx.saved_tensors
        scale = torch.abs(diff)
        scale[scale > ctx.thresh + ctx.alpha] = 1
        tag = (scale <= ctx.thresh + ctx.alpha) & (scale >= ctx.thresh)
        scale[tag] = 2 - (scale[tag] - ctx.thresh) / ctx.alpha
        tag = scale < ctx.thresh
        scale[tag] = 2 * scale[tag] / ctx.thresh
        diff[diff > 0] = 1.0
        diff[diff < 0] = -1.0
        diff = diff * scale * gradOutput / scale.numel()
        return diff, Variable(torch.Tensor([0])), None, None


class MyLossFunction(Function):
    @staticmethod
    def forward(ctx, input1, input2, upper_thresh=5, lower_thresh=1):
        ctx.upper_thresh = upper_thresh
        ctx.lower_thresh = lower_thresh
        diff = input1 - input2
        ctx.save_for_backward(diff)
        return torch.mean(torch.abs(diff))

    @staticmethod
    def backward(ctx, gradOutput):
        (diff,) = ctx.saved_tensors
        scale = torch.abs(diff)
        scale[scale > ctx.upper_thresh] = 1
        tag = (scale <= ctx.upper_thresh) & (scale >= ctx.lower_thresh)
        scale[tag] = (
            2
            - torch.abs(scale[tag] - (ctx.upper_thresh + ctx.lower_thresh) / 2.0) / 2.0
        )
        diff[diff > 0] = 1
        diff[diff < 0] = -1
        diff = diff * scale * gradOutput
        return diff, Variable(torch.Tensor([0])), None, None



import torch
import torch.nn as nn
import torch.nn.functional as F
from .gma import Aggregate


class PCBlock4_Deep_nopool_res(nn.Module):
    def __init__(self, C_in, C_out, k_conv):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = self.ffn2(x)
        return x


class SKMotionEncoder6_Deep_nopool_res(nn.Module):
    def __init__(self, corr_levels, corr_radius, k_conv):
        super().__init__()
        cor_planes = corr_levels * (2 * corr_radius + 1) ** 2
        self.convc1 = PCBlock4_Deep_nopool_res(cor_planes, 256, k_conv=k_conv)
        self.convc2 = PCBlock4_Deep_nopool_res(256, 192, k_conv=k_conv)

        self.convf1 = nn.Conv2d(2, 128, 1, 1, 0)
        self.convf2 = PCBlock4_Deep_nopool_res(128, 64, k_conv=k_conv)

        self.conv = PCBlock4_Deep_nopool_res(64 + 192, 128 - 2, k_conv=k_conv)

    def forward(self, flow, corr):
        cor = F.gelu(self.convc1(corr))

        cor = self.convc2(cor)

        flo = self.convf1(flow)
        flo = self.convf2(flo)

        cor_flo = torch.cat([cor, flo], dim=1)
        out = self.conv(cor_flo)

        return torch.cat([out, flow], dim=1)


class SKUpdateBlock6_Deep_nopoolres_AllDecoder(nn.Module):
    def __init__(
        self, corr_levels, corr_radius, k_conv, PCUpdater_conv, num_heads, hidden_dim
    ):
        super().__init__()
        self.encoder = SKMotionEncoder6_Deep_nopool_res(
            corr_levels, corr_radius, k_conv
        )
        self.gru = PCBlock4_Deep_nopool_res(
            128 + hidden_dim + hidden_dim + 128, 128, k_conv=PCUpdater_conv
        )
        self.flow_head = PCBlock4_Deep_nopool_res(128, 2, k_conv=k_conv)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

        self.aggregator = Aggregate(dim=128, dim_head=128, heads=num_heads)

    def forward(self, net, inp, corr, flow, attention):
        motion_features = self.encoder(flow, corr)
        motion_features_global = self.aggregator(attention, motion_features)
        inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)

        # Attentional update
        net = self.gru(torch.cat([net, inp_cat], dim=1))

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow



import math
import torch
import torch.nn as nn
import torch.nn.functional as F


class ResidualBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn="group", stride=1):
        super(ResidualBlock, self).__init__()

        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, padding=1, stride=stride
        )
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes)
            self.norm2 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes)
            self.norm2 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            if not stride == 1:
                self.norm3 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm3
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class BottleneckBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn="group", stride=1):
        super(BottleneckBlock, self).__init__()

        self.conv1 = nn.Conv2d(in_planes, planes // 4, kernel_size=1, padding=0)
        self.conv2 = nn.Conv2d(
            planes // 4, planes // 4, kernel_size=3, padding=1, stride=stride
        )
        self.conv3 = nn.Conv2d(planes // 4, planes, kernel_size=1, padding=0)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes // 4)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes // 4)
            self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm4 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes // 4)
            self.norm2 = nn.BatchNorm2d(planes // 4)
            self.norm3 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes // 4)
            self.norm2 = nn.InstanceNorm2d(planes // 4)
            self.norm3 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            self.norm3 = nn.Sequential()
            if not stride == 1:
                self.norm4 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm4
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))
        y = self.relu(self.norm3(self.conv3(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class BasicEncoder(nn.Module):
    def __init__(self, output_dim=128, norm_fn="batch", dropout=0.0):
        super(BasicEncoder, self).__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 64
        self.layer1 = self._make_layer(64, stride=1)
        self.layer2 = self._make_layer(96, stride=2)
        self.layer3 = self._make_layer(128, stride=2)

        # output convolution
        self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = ResidualBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = ResidualBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


class IPTHeadEncoder(nn.Module):
    """docstring for IPTHead"""

    def __init__(self, output_dim=128, norm_fn="batch", dropout=0.0):
        super(IPTHeadEncoder, self).__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        half_out_dim = max(output_dim // 2, 64)
        self.layer1 = ResidualBlock(64, half_out_dim, self.norm_fn, stride=2)
        self.layer2 = ResidualBlock(half_out_dim, output_dim, self.norm_fn, stride=2)

        # # output convolution; this can solve mixed memory warning, not know why
        # self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.relu1(self.norm1(self.conv1(x)))
        x = self.layer1(x)
        x = self.layer2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


class BasicConvEncoder(nn.Module):
    """docstring for BasicConvEncoder"""

    def __init__(self, output_dim=128, norm_fn="batch", dropout=0.0):
        super(BasicConvEncoder, self).__init__()
        self.norm_fn = norm_fn

        half_out_dim = max(output_dim // 2, 64)

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)
            self.norm2 = nn.GroupNorm(num_groups=8, num_channels=64)
            self.norm3 = nn.GroupNorm(num_groups=8, num_channels=64)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64)
            self.norm2 = nn.BatchNorm2d(half_out_dim)
            self.norm3 = nn.BatchNorm2d(output_dim)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64)
            self.norm2 = nn.InstanceNorm2d(half_out_dim)
            self.norm3 = nn.InstanceNorm2d(output_dim)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            self.norm3 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.conv2 = nn.Conv2d(64, half_out_dim, kernel_size=3, stride=2, padding=1)
        self.conv3 = nn.Conv2d(
            half_out_dim, output_dim, kernel_size=3, stride=2, padding=1
        )

        # # output convolution; this can solve mixed memory warning, not know why
        # self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = F.relu(self.norm1(self.conv1(x)), inplace=True)
        x = F.relu(self.norm2(self.conv2(x)), inplace=True)
        x = F.relu(self.norm3(self.conv3(x)), inplace=True)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


class SmallEncoder(nn.Module):
    def __init__(self, output_dim=128, norm_fn="batch", dropout=0.0):
        super(SmallEncoder, self).__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=32)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(32)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(32)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 32, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 32
        self.layer1 = self._make_layer(32, stride=1)
        self.layer2 = self._make_layer(64, stride=2)
        self.layer3 = self._make_layer(96, stride=2)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        self.conv2 = nn.Conv2d(96, output_dim, kernel_size=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = BottleneckBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = BottleneckBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


class MultiHeadAttention(nn.Module):
    """Multi-Head Attention module"""

    def __init__(self, d_model, n_head, dropout=0.1):
        super().__init__()

        assert d_model % n_head == 0

        self.n_head = n_head
        self.d_model = d_model
        self.d_head = self.d_model // self.n_head

        self.w_qs = nn.Linear(
            self.d_model, self.d_model, bias=False
        )  # TODO: enable bias
        self.w_ks = nn.Linear(self.d_model, self.d_model, bias=False)
        self.w_vs = nn.Linear(self.d_model, self.d_model, bias=False)
        self.fc = nn.Linear(self.d_model, self.d_model)

        # self.attention = ScaledDotProductAttention(temperature=d_k ** 0.5) # TODO

        self.dropout = nn.Dropout(dropout)
        # self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)

    def forward(self, q, k, v):
        """
        q: shape of N*len*C
        """
        d_head, n_head = self.d_head, self.n_head
        sz_b, len_q, len_k, len_v = q.size(0), q.size(1), k.size(1), v.size(1)

        # residual = q

        # Pass through the pre-attention projection: b x lq x (n*dv)
        # Separate different heads: b x lq x n x dv
        q = self.w_qs(q).view(sz_b, len_q, n_head, d_head)
        k = self.w_ks(k).view(sz_b, len_k, n_head, d_head)
        v = self.w_vs(v).view(sz_b, len_v, n_head, d_head)

        # Transpose for attention dot product: b x n x lq x dv
        q, k, v = q.transpose(1, 2), k.transpose(1, 2), v.transpose(1, 2)

        attn = torch.matmul(q, k.transpose(-1, -2)) / math.sqrt(self.d_head)
        attn = self.dropout(F.softmax(attn, dim=-1))
        q_updated = torch.matmul(attn, v)

        # Transpose to move the head dimension back: b x lq x n x dv
        # Combine the last two dimensions to concatenate all the heads together: b x lq x (n*dv)
        q_updated = q_updated.transpose(1, 2).contiguous().view(sz_b, len_q, -1)
        q_updated = self.dropout(self.fc(q_updated))
        # q_updated += residual

        # q_updated = self.layer_norm(q_updated)

        return q_updated, attn


class AnchorEncoderBlock(nn.Module):
    def __init__(self, anchor_dist, d_model, num_heads, d_ff, dropout=0.0):
        super().__init__()

        self.anchor_dist = anchor_dist
        self.half_anchor_dist = anchor_dist // 2

        self.selfAttn = MultiHeadAttention(d_model, num_heads, dropout)
        self.dropout = nn.Dropout(dropout)
        self.layer_norm_1 = nn.LayerNorm(d_model)

        self.FFN = nn.Sequential(
            nn.Linear(d_model, d_ff),
            nn.ReLU(),
            nn.Linear(d_ff, d_model),
        )

        self.layer_norm_2 = nn.LayerNorm(d_model)

    def forward(self, inputs):
        """
        inputs: batches with N*C*H*W
        """
        N, C, H, W = inputs.shape

        x = inputs
        anchors = inputs[
            :,
            :,
            self.half_anchor_dist :: self.anchor_dist,
            self.half_anchor_dist :: self.anchor_dist,
        ].clone()

        # flatten feature maps
        x = x.reshape(N, C, H * W).transpose(-1, -2)
        anchors = anchors.reshape(N, C, anchors.shape[2] * anchors.shape[3]).transpose(
            -1, -2
        )

        # two-stage multi-head self-attention
        anchors_new = self.dropout(self.selfAttn(anchors, x, x)[0])
        residual = self.dropout(self.selfAttn(x, anchors_new, anchors_new)[0])

        norm_1 = self.layer_norm_1(x + residual)
        x_linear = self.dropout(self.FFN(norm_1))
        x_new = self.layer_norm_2(norm_1 + x_linear)

        outputs = x_new.transpose(-1, -2).reshape(N, C, H, W)
        return outputs


class EncoderBlock(nn.Module):
    def __init__(self, d_model, num_heads, d_ff, dropout=0.0):
        super().__init__()

        self.selfAttn = MultiHeadAttention(d_model, num_heads, dropout)
        self.dropout = nn.Dropout(dropout)
        self.layer_norm_1 = nn.LayerNorm(d_model)

        self.FFN = nn.Sequential(
            nn.Linear(d_model, d_ff),
            nn.ReLU(),
            nn.Linear(d_ff, d_model),
        )

        self.layer_norm_2 = nn.LayerNorm(d_model)

    def forward(self, x):
        """
        x: input batches with N*C*H*W
        """
        N, C, H, W = x.shape

        # update x
        x = x.reshape(N, C, H * W).transpose(-1, -2)

        residual = self.dropout(self.selfAttn(x, x, x)[0])
        norm_1 = self.layer_norm_1(x + residual)
        x_linear = self.dropout(self.FFN(norm_1))
        x_new = self.layer_norm_2(norm_1 + x_linear)

        outputs = x_new.transpose(-1, -2).reshape(N, C, H, W)
        return outputs


class ReduceEncoderBlock(nn.Module):
    def __init__(self, d_model, num_heads, d_ff, dropout=0.0):
        super().__init__()

        self.reduce = nn.Sequential(
            nn.Conv2d(d_model, d_model, 2, 2), nn.Conv2d(d_model, d_model, 2, 2)
        )
        # self.reduce = nn.Sequential(
        #       nn.AvgPool2d(16, 16)
        #     )

        self.selfAttn = MultiHeadAttention(d_model, num_heads, dropout)
        self.dropout = nn.Dropout(dropout)
        self.layer_norm_1 = nn.LayerNorm(d_model)

        self.FFN = nn.Sequential(
            nn.Linear(d_model, d_ff), nn.ReLU(), nn.Linear(d_ff, d_model)
        )

        self.layer_norm_2 = nn.LayerNorm(d_model)

    def forward(self, x):
        """
        x: input batches with N*C*H*W
        """
        N, C, H, W = x.shape
        x_reduced = self.reduce(x)

        # update x
        x = x.reshape(N, C, H * W).transpose(-1, -2)
        x_reduced = x_reduced.reshape(N, C, -1).transpose(-1, -2)

        # print('x ', x.shape)
        # print('x_reduced ', x_reduced.shape)
        # exit()

        residual = self.dropout(self.selfAttn(x, x_reduced, x_reduced)[0])

        norm_1 = self.layer_norm_1(x + residual)
        x_linear = self.dropout(self.FFN(norm_1))
        x_new = self.layer_norm_2(norm_1 + x_linear)

        outputs = x_new.transpose(-1, -2).reshape(N, C, H, W)
        return outputs


def window_partition(x, window_size):
    """
    Args:
        x: (B, H, W, C)
        window_size (int): window size

    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """
    B, H, W, C = x.shape
    x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
    windows = (
        x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
    )
    return windows


def window_reverse(windows, window_size, H, W):
    """
    Args:
        windows: (num_windows*B, window_size, window_size, C)
        window_size (int): Window size
        H (int): Height of image
        W (int): Width of image

    Returns:
        x: (B, H, W, C)
    """
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.view(
        B, H // window_size, W // window_size, window_size, window_size, -1
    )
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x


class LayerEncoderBlock(nn.Module):
    def __init__(self, win_size, d_model, num_heads, d_ff, dropout=0.0):
        super().__init__()

        self.win_size = win_size
        self.down_factor = 4
        self.unfold_stride = int(self.win_size // self.down_factor)

        self.stride_list = [
            math.floor(win_size / self.down_factor**idx)
            for idx in range(8)
            if win_size / self.down_factor**idx >= 1
        ]
        # [16, 4, 1]

        self.reduce = nn.Sequential(nn.AvgPool2d(self.down_factor, self.down_factor))

        self.selfAttn = MultiHeadAttention(d_model, num_heads, dropout)
        self.crossAttn = MultiHeadAttention(d_model, num_heads, dropout)

        self.dropout = nn.Dropout(dropout)
        self.layerNormSelf = nn.LayerNorm(d_model)
        self.layerNormCross = nn.LayerNorm(d_model)

        self.FFN = nn.Sequential(
            nn.Linear(d_model, d_ff), nn.ReLU(), nn.Linear(d_ff, d_model)
        )

        self.layer_norm_out = nn.LayerNorm(d_model)

    def Circular_pad2D(self, x, pad_right, pad_bottom):
        """
        x: (N, H, W, C)
        x_pad: (N, H_pad, W_pad, C)
        """
        N, H, W, C = x.shape

        H_pad = H + pad_bottom
        W_pad = W + pad_right

        H_repeat = math.ceil(H_pad / H)
        W_repeat = math.ceil(W_pad / W)
        x_repeat = x.repeat(1, H_repeat, W_repeat, 1)

        x_pad = x_repeat[:, :H_pad, :W_pad, :]
        return x_pad

    def pad_fit_win(self, x, win_size):
        N, H, W, C = x.shape

        W_ = math.ceil(W / win_size) * win_size
        H_ = math.ceil(H / win_size) * win_size
        padRight = W_ - W
        padBottom = H_ - H

        x_pad = self.Circular_pad2D(x, padRight, padBottom)  # N*H_*W_*C
        return x_pad

    def self_attention(self, x):
        """
        x: (N, H, W, C)
        out: (N, H, W, C)
        """
        N, H, W, C = x.shape
        x_pad = self.pad_fit_win(x, self.win_size)  # N*H_*W_*C
        _, H_, W_, _ = x_pad.shape

        # x_pad = F.pad(x.permute(xxx), (0, padRight, 0, padBottom), mode='reflect') # N*C*H_*W_

        x_window = window_partition(
            x_pad, self.win_size
        )  # (num_win*B, win_size, win_size, C)
        x_window = x_window.view(
            -1, self.win_size * self.win_size, C
        )  # (num_win*B, win_size*win_size, C)

        # self-attention
        residual = self.dropout(self.selfAttn(x_window, x_window, x_window)[0])
        residual = residual.view(-1, self.win_size, self.win_size, C)
        residual = window_reverse(residual, self.win_size, H_, W_)  # (N, H_, W_, C)

        out = x_pad + residual
        out = out[:, :H, :W, :]
        return out

    def cross_attention(self, query, keyVal):
        """
        query: (N, qH, qW, C)
        keyVal: (N, kH, kW, C)
        out: (N, qH, qW, C)
        """
        _, qH, qW, C = query.shape
        _, kH, kW, C = keyVal.shape

        # print('in query ', query.shape)
        # print('in keyVal ', keyVal.shape)
        # print('-')

        query = self.pad_fit_win(query, self.win_size)  # N*H_*W_*C
        _, qH_, qW_, C = query.shape

        query_win = window_partition(query, self.win_size)
        query_win = query_win.view(
            -1, self.win_size * self.win_size, C
        )  # (num_win*B, win_size*win_size, C)

        # pad and unfold keyVal
        kW_ = (
            math.ceil(kW / self.unfold_stride) - 1
        ) * self.unfold_stride + self.win_size
        kH_ = (
            math.ceil(kH / self.unfold_stride) - 1
        ) * self.unfold_stride + self.win_size
        padRight = kW_ - kW
        padBottom = kH_ - kH

        keyVal_pad = self.Circular_pad2D(keyVal, padRight, padBottom)
        keyVal = F.unfold(
            keyVal_pad.permute(0, 3, 1, 2), self.win_size, stride=self.unfold_stride
        )  # (N, C*win_size*win_size, num_win)
        keyVal = (
            keyVal.permute(0, 2, 1)
            .reshape(-1, C, self.win_size * self.win_size)
            .permute(0, 2, 1)
        )  # (num_win*B, win_size*win_size, C)

        # print('win query ', query_win.shape)
        # print('win keyVal ', keyVal.shape)
        # print('-')

        residual = self.dropout(self.crossAttn(query_win, keyVal, keyVal)[0])
        residual = residual.view(-1, self.win_size, self.win_size, C)
        residual = window_reverse(residual, self.win_size, qH_, qW_)  # (N, H, W, C)

        out = query + residual
        out = out[:, :qH, :qW, :]
        return out

    def forward(self, x):
        """
        x: input batches with N*C*H*W
        """
        N, C, H, W = x.shape
        x = x.permute(0, 2, 3, 1)  # N*H*W*C
        x = self.pad_fit_win(x, self.win_size)  # pad

        # layered self-attention
        layerAttnList = []
        strideListLen = len(self.stride_list)
        for idx in range(strideListLen):
            x_attn = self.self_attention(x)  # built-in shortcut
            x_attn = self.layerNormSelf(x_attn)
            layerAttnList.append(x_attn)

            if idx < strideListLen - 1:
                x = self.reduce(x_attn.permute(0, 3, 1, 2))  # N*C*H*W
                x = x.permute(0, 2, 3, 1)  # N*H*W*C

        # layered cross-attention
        KeyVal = layerAttnList[-1]
        for idx in range(strideListLen - 1, 0, -1):
            Query = layerAttnList[idx - 1]
            Query = self.cross_attention(Query, KeyVal)  # built-in shortcut
            Query = self.layerNormCross(Query)

            KeyVal = Query

        Query = Query[:, :H, :W, :]  # unpad

        q_residual = self.dropout(self.FFN(Query))
        x_new = self.layer_norm_out(Query + q_residual)

        outputs = x_new.permute(0, 3, 1, 2)
        return outputs


class BasicLayerEncoderBlock(nn.Module):
    def __init__(self, win_size, d_model, num_heads, d_ff, dropout=0.0):
        super().__init__()

        self.win_size = win_size
        self.down_factor = 2
        self.unfold_stride = int(self.win_size // self.down_factor)

        self.stride_list = [
            math.floor(win_size / self.down_factor**idx)
            for idx in range(8)
            if win_size / self.down_factor**idx >= 1
        ]
        # [16, 8, 4, 2, 1]

        self.reduce = nn.Sequential(nn.AvgPool2d(self.down_factor, self.down_factor))

        self.selfAttn = MultiHeadAttention(d_model, num_heads, dropout)
        self.crossAttn = MultiHeadAttention(d_model, num_heads, dropout)

        self.dropout = nn.Dropout(dropout)
        self.layerNormSelf = nn.LayerNorm(d_model)
        self.layerNormCross = nn.LayerNorm(d_model)

        self.FFN = nn.Sequential(
            nn.Linear(d_model, d_ff), nn.ReLU(), nn.Linear(d_ff, d_model)
        )

        self.layer_norm_out = nn.LayerNorm(d_model)

    def Circular_pad2D(self, x, pad_right, pad_bottom):
        """
        x: (N, H, W, C)
        x_pad: (N, H_pad, W_pad, C)
        """
        N, H, W, C = x.shape

        H_pad = H + pad_bottom
        W_pad = W + pad_right

        H_repeat = math.ceil(H_pad / H)
        W_repeat = math.ceil(W_pad / W)
        x_repeat = x.repeat(1, H_repeat, W_repeat, 1)

        x_pad = x_repeat[:, :H_pad, :W_pad, :]
        return x_pad

    def pad_fit_win(self, x, win_size):
        N, H, W, C = x.shape

        W_ = math.ceil(W / win_size) * win_size
        H_ = math.ceil(H / win_size) * win_size
        padRight = W_ - W
        padBottom = H_ - H

        x_pad = self.Circular_pad2D(x, padRight, padBottom)  # N*H_*W_*C
        return x_pad

    def self_attention(self, x):
        """
        x: (N, H, W, C)
        out: (N, H, W, C)
        """
        N, H, W, C = x.shape
        x_pad = self.pad_fit_win(x, self.win_size)  # N*H_*W_*C
        _, H_, W_, _ = x_pad.shape

        # x_pad = F.pad(x.permute(xxx), (0, padRight, 0, padBottom), mode='reflect') # N*C*H_*W_

        x_window = window_partition(
            x_pad, self.win_size
        )  # (num_win*B, win_size, win_size, C)
        x_window = x_window.view(
            -1, self.win_size * self.win_size, C
        )  # (num_win*B, win_size*win_size, C)

        # self-attention
        residual = self.dropout(self.selfAttn(x_window, x_window, x_window)[0])
        residual = residual.view(-1, self.win_size, self.win_size, C)
        residual = window_reverse(residual, self.win_size, H_, W_)  # (N, H_, W_, C)

        out = x_pad + residual
        out = out[:, :H, :W, :]
        return out

    def cross_attention(self, query, keyVal, query_win_size):
        """
        query: (N, qH, qW, C)
        keyVal: (N, kH, kW, C)
        out: (N, qH, qW, C)
        """
        _, qH, qW, C = query.shape

        query_win = window_partition(query, query_win_size)
        query_win = query_win.view(
            -1, query_win_size * query_win_size, C
        )  # (num_win*B, win_size*win_size, C)

        keyWinSize = query_win_size // 2
        keyVal_win = window_partition(keyVal, keyWinSize)
        keyVal_win = keyVal_win.view(
            -1, keyWinSize * keyWinSize, C
        )  # (num_win*B, win_size*win_size, C)

        residual = self.dropout(self.crossAttn(query_win, keyVal_win, keyVal_win)[0])
        residual = residual.view(-1, query_win_size, query_win_size, C)
        residual = window_reverse(residual, query_win_size, qH, qW)  # (N, H, W, C)

        out = query + residual
        return out

    def forward(self, x):
        """
        x: input batches with N*C*H*W
        """
        N, C, H, W = x.shape
        x = x.permute(0, 2, 3, 1)  # N*H*W*C
        x = self.pad_fit_win(x, self.win_size)  # pad

        # layered self-attention
        layerAttnList = []
        strideListLen = len(self.stride_list)
        for idx in range(strideListLen):
            x_attn = self.self_attention(x)  # built-in shortcut
            x_attn = self.layerNormSelf(x_attn)
            layerAttnList.append(x_attn)

            if idx < strideListLen - 1:
                x = self.reduce(x_attn.permute(0, 3, 1, 2))  # N*C*H*W
                x = x.permute(0, 2, 3, 1)  # N*H*W*C

        # layered cross-attention
        KeyVal = layerAttnList[-1]
        for idx in range(strideListLen - 1, 0, -1):
            Query = layerAttnList[idx - 1]
            QueryWinSize = self.stride_list[idx - 1]

            Query = self.cross_attention(
                Query, KeyVal, QueryWinSize
            )  # built-in shortcut
            Query = self.layerNormCross(Query)

            KeyVal = Query

        Query = Query[:, :H, :W, :]  # unpad

        q_residual = self.dropout(self.FFN(Query))
        x_new = self.layer_norm_out(Query + q_residual)

        outputs = x_new.permute(0, 3, 1, 2)
        return outputs


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.0):
        super().__init__()

        self.max_len = 256
        self.d_model = d_model

        self._update_PE_table(self.max_len, self.d_model // 2)

    def _update_PE_table(self, max_len, d_model):
        self.PE_table = torch.zeros(max_len, d_model)

        pos = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        denominator = torch.pow(10000, torch.arange(0, d_model, 2).float() / d_model)

        self.PE_table[:, 0::2] = torch.sin(pos / denominator)
        self.PE_table[:, 1::2] = torch.cos(pos / denominator)

    def forward(self, x):
        """x: image batches with N*C*H*W"""

        N, C, H, W = x.shape
        max_hw = max(H, W)

        if max_hw > self.max_len or self.d_model != C:
            self.max_len = max_hw
            self.d_model = C

            self._update_PE_table(self.max_len, self.d_model // 2)

        if self.PE_table.dtype != x.dtype:
            self.PE_table = self.PE_table.to(dtype=x.dtype)
        if self.PE_table.device != x.device:
            self.PE_table = self.PE_table.to(x.device)

        h_pos_emb = self.PE_table[:H, :].unsqueeze(1).repeat(1, W, 1)  # H*W*C/2
        w_pos_emb = self.PE_table[:W, :].unsqueeze(0).repeat(H, 1, 1)  # H*W*C/2
        pos_emb = (
            torch.cat([h_pos_emb, w_pos_emb], dim=-1)
            .permute([2, 0, 1])
            .unsqueeze(0)
            .repeat(N, 1, 1, 1)
        )  # N*C*H*W

        output = x + pos_emb
        return output


class TransformerEncoder(nn.Module):
    def __init__(self, anchor_dist, num_blocks, d_model, num_heads, d_ff, dropout=0.0):
        super().__init__()

        self.anchor_dist = anchor_dist

        blocks_list = []
        for idx in range(num_blocks):
            # blocks_list.append( AnchorEncoderBlock(anchor_dist, d_model, num_heads, d_ff, dropout) )
            # blocks_list.append( EncoderBlock(d_model, num_heads, d_ff, dropout) )
            blocks_list.append(ReduceEncoderBlock(d_model, num_heads, d_ff, dropout))
        #   blocks_list.append( BasicLayerEncoderBlock(anchor_dist, d_model, num_heads, d_ff, dropout) )

        self.blocks = nn.Sequential(*blocks_list)

        self.posEmbedding = PositionalEncoding(d_model, dropout)

    def forward(self, x):
        x_w_pos = self.posEmbedding(x)
        x_updated = self.blocks(x_w_pos)

        return x_updated


class RawInputTransEncoder(nn.Module):
    def __init__(self, anchor_dist, num_blocks, d_model, num_heads, d_ff, dropout=0.0):
        super().__init__()

        self.anchor_dist = anchor_dist

        self.linear = nn.Conv2d(3, d_model, 8, 8)

        blocks_list = []
        for idx in range(num_blocks):
            # blocks_list.append( AnchorEncoderBlock(anchor_dist, d_model, num_heads, d_ff, dropout) )
            # blocks_list.append( EncoderBlock(d_model, num_heads, d_ff, dropout) )
            # blocks_list.append( ReduceEncoderBlock(d_model, num_heads, d_ff, dropout) )
            blocks_list.append(
                LayerEncoderBlock(anchor_dist, d_model, num_heads, d_ff, dropout)
            )

        self.blocks = nn.Sequential(*blocks_list)

        self.posEmbedding = PositionalEncoding(d_model, dropout)

        # initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(
                m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm, nn.LayerNorm)
            ):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.linear(x)
        x_w_pos = self.posEmbedding(x)
        x_updated = self.blocks(x_w_pos)

        if is_list:
            x_updated = torch.split(x_updated, [batch_dim, batch_dim], dim=0)

        return x_updated


class GlobalLocalBlock(nn.Module):
    def __init__(self, anchor_dist, d_model, num_heads, out_dim, dropout=0.0, stride=1):
        super().__init__()

        self.anchor_dist = anchor_dist
        self.half_anchor_dist = anchor_dist // 2
        self.d_model = d_model
        self.out_dim = out_dim

        self.selfAttn = MultiHeadAttention(d_model, num_heads, dropout)
        self.dropout = nn.Dropout(dropout)
        self.layer_norm_1 = nn.LayerNorm(d_model)

        self.resBlock_1 = ResidualBlock(
            d_model, d_model, norm_fn="instance", stride=stride
        )
        self.change_channel = nn.Linear(d_model, out_dim)
        self.resBlock_2 = ResidualBlock(out_dim, out_dim, norm_fn="instance", stride=1)

        self.posEmbedding = PositionalEncoding(d_model, dropout)

    def forward(self, inputs):
        """
        inputs: batches with N*H*W*C
        """

        # local update 1
        x = self.resBlock_1(inputs)
        x = self.posEmbedding(x)
        anchors = x[
            :,
            :,
            self.half_anchor_dist :: self.anchor_dist,
            self.half_anchor_dist :: self.anchor_dist,
        ].clone()

        # flatten feature maps
        N, C, H, W = x.shape
        x = x.reshape(N, C, H * W).transpose(-1, -2)
        anchors = anchors.reshape(N, C, anchors.shape[2] * anchors.shape[3]).transpose(
            -1, -2
        )

        # gloabl update with two-stage multi-head self-attention
        anchors_new = self.dropout(self.selfAttn(anchors, x, x)[0])
        residual = self.dropout(self.selfAttn(x, anchors_new, anchors_new)[0])
        norm_1 = self.layer_norm_1(x + residual)

        # local update 2
        norm_1 = self.change_channel(norm_1)
        norm_1 = norm_1.transpose(-1, -2).reshape(N, self.out_dim, H, W)
        outputs = self.resBlock_2(norm_1)

        return outputs


class GlobalLocalEncoder(nn.Module):
    def __init__(self, anchor_dist, output_dim, dropout=0.0):
        super().__init__()

        self.anchor_dist = anchor_dist
        self.output_dim = output_dim

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.norm1 = nn.InstanceNorm2d(64)
        self.relu1 = nn.ReLU(inplace=True)

        self.layer1 = GlobalLocalBlock(self.anchor_dist, 64, 2, 96, dropout, stride=2)
        self.layer2 = GlobalLocalBlock(self.anchor_dist, 96, 3, 96, dropout, stride=1)
        self.layer3 = GlobalLocalBlock(
            self.anchor_dist // 2, 96, 4, 128, dropout, stride=2
        )

        # output convolution
        self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        # initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(
                m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm, nn.LayerNorm)
            ):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    # def _make_layer(self, in_dim, out_dim, dropout=0., stride=1):
    #     layer1 = GlobalLocalBlock(self.anchor_dist, in_dim, in_dim//32, out_dim, dropout=0., stride=stride)
    #     layer2 = GlobalLocalBlock(self.anchor_dist, out_dim, out_dim//32, out_dim, dropout=0., stride=1)
    #     layers = (layer1, layer2)

    #     return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.relu1(self.norm1(self.conv1(x)))

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x



"""Validate optical flow estimation performance on standard datasets."""

# =============================================================================
# Copyright 2024 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from argparse import ArgumentParser
from pathlib import Path
import sys

import cv2 as cv
import numpy as np
import onnx
import onnxruntime

this_dir = Path(__file__).parent.resolve()
sys.path.insert(0, str(this_dir.parent.parent.parent))

from ptlflow.utils import flow_utils


def _init_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "onnx_path",
        type=str,
        help="Path to the ONNX model.",
    )
    parser.add_argument(
        "--image_paths",
        type=str,
        nargs=2,
        default=(
            str(this_dir / "image_samples" / "000000_10.png"),
            str(this_dir / "image_samples" / "000000_11.png"),
        ),
        help="Path to two images to estimate the optical flow with the ONNX model.",
    )
    parser.add_argument(
        "--output_path",
        type=str,
        default=".",
        help="Path to the directory where the output flow will be saved.",
    )
    parser.add_argument(
        "--input_size",
        type=int,
        nargs=2,
        default=(384, 1280),
        help="Size of the input image.",
    )
    return parser


def load_images(image_paths):
    images = [cv.imread(p) for p in image_paths]
    images = [cv.resize(im, args.input_size[::-1]) for im in images]
    images = np.stack(images)
    images = images.transpose(0, 3, 1, 2)[None]
    images = images.astype(np.float32) / 255.0
    return images


if __name__ == "__main__":
    parser = _init_parser()
    args = parser.parse_args()

    onnx_model = onnx.load(args.onnx_path)
    # onnx.checker.check_model(onnx_model)

    inputs = load_images(args.image_paths)
    print(inputs.shape, inputs.max())

    ort_session = onnxruntime.InferenceSession(
        args.onnx_path, providers=["CUDAExecutionProvider"]
    )

    ort_inputs = {ort_session.get_inputs()[0].name: inputs}
    ort_outs = ort_session.run(None, ort_inputs)
    flow_pred = ort_outs[0][0].transpose(1, 2, 0)
    print(flow_pred.shape, flow_pred.min(), flow_pred.max())

    output_dir = Path(args.output_path)
    output_dir.mkdir(parents=True, exist_ok=True)

    flo_output_path = output_dir / f"flow_pred.flo"
    flow_utils.flow_write(flo_output_path, flow_pred)
    print(f"Saved flow prediction to: {flo_output_path}")

    viz_output_path = output_dir / f"flow_pred_viz.png"
    flow_viz = flow_utils.flow_to_rgb(flow_pred)
    cv.imwrite(str(viz_output_path), cv.cvtColor(flow_viz, cv.COLOR_RGB2BGR))
    print(f"Saved flow prediction visualization to: {viz_output_path}")



import torch
from torch import nn, einsum
from einops import rearrange
import torch.nn.functional as F
import math


# The position encoding block is from https://github.com/haofeixu/gmflow
class PositionEmbeddingSine(nn.Module):
    """
    This is a more standard version of the position embedding, very similar to the one
    used by the Attention is all you need paper, generalized to work on images.
    """

    def __init__(self, num_pos_feats=64, temperature=10000, normalize=True, scale=None):
        super().__init__()
        self.num_pos_feats = num_pos_feats
        self.temperature = temperature
        self.normalize = normalize
        if scale is not None and normalize is False:
            raise ValueError("normalize should be True if scale is passed")
        if scale is None:
            scale = 2 * math.pi
        self.scale = scale

    def forward(self, x):
        # x = tensor_list.tensors  # [B, C, H, W]
        # mask = tensor_list.mask  # [B, H, W], input with padding, valid as 0
        b, c, h, w = x.size()
        mask = torch.ones((b, h, w), device=x.device)  # [B, H, W]
        y_embed = mask.cumsum(1, dtype=x.dtype)
        x_embed = mask.cumsum(2, dtype=x.dtype)
        if self.normalize:
            eps = 1e-6
            y_embed = y_embed / (y_embed[:, -1:, :] + eps) * self.scale
            x_embed = x_embed / (x_embed[:, :, -1:] + eps) * self.scale

        dim_t = torch.arange(self.num_pos_feats, dtype=x.dtype, device=x.device)
        dim_t = self.temperature ** (2 * (dim_t // 2) / self.num_pos_feats)

        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        pos_x = torch.stack(
            (pos_x[:, :, :, 0::2].sin(), pos_x[:, :, :, 1::2].cos()), dim=4
        ).flatten(3)
        pos_y = torch.stack(
            (pos_y[:, :, :, 0::2].sin(), pos_y[:, :, :, 1::2].cos()), dim=4
        ).flatten(3)
        pos = torch.cat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)
        return pos


def PatchExtra(x, size):
    B, C, H, W = x.shape
    x = F.pad(x, (size // 2, size // 2, size // 2, size // 2), "replicate")
    return F.unfold(x, [size, size], stride=1, padding=0).reshape(
        B, C, size * size, H, W
    )


def ImgShift(x, size):
    B, C, H, W = x.shape
    x = F.pad(x, (size // 2, size // 2, size // 2, size // 2), "replicate")
    return (
        F.unfold(x, [H, W], stride=1, padding=0)
        .reshape(B, C, H, W, size**2)
        .permute(4, 0, 1, 2, 3)
    )


class GlobalSimilar(nn.Module):
    def __init__(
        self,
        dim=128,
        heads=1,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.scale = (dim // heads) ** -0.5

        self.to_qk = nn.Conv2d(dim, dim * 2, 1, bias=False)

    def forward(self, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        q, k = self.to_qk(fmap).chunk(2, dim=1)
        q, k = map(lambda t: rearrange(t, "b (h d) x y -> b h x y d", h=heads), (q, k))
        q = self.scale * q
        sim = einsum("b h x y d, b h u v d -> b h x y u v", q, k)
        sim = rearrange(sim, "b h x y u v -> b h (x y) (u v)")
        attn = sim.softmax(dim=-1)

        return attn


class LocalSimilar(nn.Module):
    def __init__(
        self,
        dim=128,
        heads=1,
        size=5,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.scale = (dim // heads) ** -0.5
        self.size = size

        self.to_qk = nn.Conv2d(dim, dim * 2, 1, bias=False)

    def forward(self, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        # pos_enc = PositionEmbeddingSine(num_pos_feats=c//2)
        # fmap = fmap + pos_enc(fmap)

        q, k = self.to_qk(fmap).chunk(2, dim=1)  # B, C, H, W
        k = PatchExtra(k, self.size)  # B, C, L, H, W

        q = rearrange(q, "b (h d) x y -> b h x y d", h=heads)
        k = rearrange(k, "b (h d) l x y -> b h x y d l", h=heads)
        q = self.scale * q

        sim = einsum("b h x y d, b h x y d l -> b h x y l", q, k)
        sim = rearrange(sim, "b h x y l -> b h (x y) l")
        attn = sim.softmax(dim=-1)

        return attn


class GSA(nn.Module):
    def __init__(
        self,
        dim=128,
        heads=1,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.scale = (dim // heads) ** -0.5

        self.to_v = nn.Conv2d(dim, dim, 1, bias=False)
        self.gamma = nn.Parameter(torch.zeros(1))

    def forward(self, attn, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        v = self.to_v(fmap)
        v = rearrange(v, "b (h d) x y -> b h (x y) d", h=heads)
        out = einsum("b h i j, b h j d -> b h i d", attn, v)
        out = rearrange(out, "b h (x y) d -> b (h d) x y", x=h, y=w)

        return fmap + self.gamma * out


class ShiftLSA(nn.Module):
    def __init__(
        self,
        dim=128,
        heads=1,
        size=5,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.scale = (dim // heads) ** -0.5
        self.size = size

        self.to_f1 = nn.Conv2d(dim, dim, 1, bias=False)
        self.to_f2 = nn.Conv2d(dim, dim, 1, bias=False)

    def forward(self, attn, fmap, fmap2):
        heads, b, c, h, w = self.heads, *fmap.shape

        f1 = self.to_f1(fmap)
        f1s = PatchExtra(f1, self.size)  # b c l h w
        f1s = rearrange(f1s, "b (h d) l x y -> b h (x y) d l", h=heads)
        f1s = einsum("b h n l, b h n d l -> b h n d l", attn, f1s)

        f2 = self.to_f2(fmap2)
        f2s = ImgShift(f2, self.size)
        f2s = rearrange(f2s, "l b (h d) x y -> b h (x y) d l", h=heads)
        corr = einsum("b h n d l, b h m d l -> b h n m", f1s, f2s)
        corr = rearrange(corr, "b h n m -> b n h m").view(b, h, w, heads, h, w)

        return corr / torch.sqrt(torch.tensor(self.dim // self.heads))


class LSA(nn.Module):
    def __init__(self, dim=128, heads=1, size=5):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.scale = (dim // heads) ** -0.5

        self.to_v = nn.Conv2d(dim, dim, 1, bias=False)
        self.gamma = nn.Parameter(torch.zeros(1))
        self.size = size
        # self.bias = nn.Parameter(torch.ones(5,5))

    def forward(self, attn, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        # pos_enc = PositionEmbeddingSine(num_pos_feats=c//2)
        # position = pos_enc(fmap)
        # fmap = fmap + position

        v = self.to_v(fmap)
        v = PatchExtra(v, self.size)  # b c l h w
        v = rearrange(v, "b (h d) l x y -> b h (x y) d l", h=heads)
        out = einsum("b h n l, b h n d l -> b h n d", attn, v)
        out = rearrange(out, "b h (x y) d -> b (h d) x y", x=h, y=w)

        out = fmap + self.gamma * out
        return out


if __name__ == "__main__":
    att = GlobalSimilar(dim=128, heads=1)
    fmap = torch.randn(2, 128, 40, 90)
    out = att(fmap)

    print(out.shape)



import math

import torch
import torch.nn.functional as F
import numpy as np
from scipy import interpolate


class InputPadder:
    """Pads images such that dimensions are divisible by 8"""

    def __init__(self, dims, pad_mode, mode="sintel"):
        self.ht, self.wd = dims[-2:]
        self.pad_mode = pad_mode
        pad_ht = (((self.ht // 8) + 1) * 8 - self.ht) % 8
        pad_wd = (((self.wd // 8) + 1) * 8 - self.wd) % 8
        if mode == "sintel":
            self._pad = [
                pad_wd // 2,
                pad_wd - pad_wd // 2,
                pad_ht // 2,
                pad_ht - pad_ht // 2,
            ]
        elif mode == "kitti432":
            self._pad = [0, 0, 0, 432 - self.ht]
        elif mode == "kitti400":
            self._pad = [0, 0, 0, 400 - self.ht]
        elif mode == "kitti376":
            self._pad = [0, 0, 0, 376 - self.ht]
        else:
            self._pad = [pad_wd // 2, pad_wd - pad_wd // 2, 0, pad_ht]

    def pad(self, *inputs):
        return [F.pad(x, self._pad, mode=self.pad_mode, value=0.0) for x in inputs]

    def unpad(self, x):
        ht, wd = x.shape[-2:]
        c = [self._pad[2], ht - self._pad[3], self._pad[0], wd - self._pad[1]]
        return x[..., c[0] : c[1], c[2] : c[3]]


def forward_interpolate(flow):
    dtype = flow.dtype
    flow = flow.detach().cpu().numpy()
    dx, dy = flow[0], flow[1]

    ht, wd = dx.shape
    x0, y0 = np.meshgrid(np.arange(wd), np.arange(ht))

    x1 = x0 + dx
    y1 = y0 + dy

    x1 = x1.reshape(-1)
    y1 = y1.reshape(-1)
    dx = dx.reshape(-1)
    dy = dy.reshape(-1)

    valid = (x1 > 0) & (x1 < wd) & (y1 > 0) & (y1 < ht)
    x1 = x1[valid]
    y1 = y1[valid]
    dx = dx[valid]
    dy = dy[valid]

    flow_x = interpolate.griddata(
        (x1, y1), dx, (x0, y0), method="nearest", fill_value=0
    )

    flow_y = interpolate.griddata(
        (x1, y1), dy, (x0, y0), method="nearest", fill_value=0
    )

    flow = np.stack([flow_x, flow_y], axis=0)
    return torch.from_numpy(flow).to(dtype=dtype)


def bilinear_sampler(img, coords, mode="bilinear", mask=False):
    """Wrapper for grid_sample, uses pixel coordinates"""
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, align_corners=True)

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.to(dtype=coords.dtype)

    return img


def indexing(img, coords, mask=False):
    """Wrapper for grid_sample, uses pixel coordinates"""
    """
        TODO: directly indexing features instead of sampling
    """
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, align_corners=True, mode="nearest")

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.to(dtype=coords.dtype)

    return img


def coords_grid(batch, ht, wd, dtype, device):
    coords = torch.meshgrid(
        torch.arange(ht, dtype=dtype, device=device),
        torch.arange(wd, dtype=dtype, device=device),
        indexing="ij",
    )
    coords = torch.stack(coords[::-1], dim=0)
    return coords[None].repeat(batch, 1, 1, 1)


def upflow8(flow, mode="bilinear"):
    new_size = (8 * flow.shape[2], 8 * flow.shape[3])
    return 8 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)


def compute_grid_indices(image_shape, patch_size, min_overlap=20):
    if min_overlap >= patch_size[0] or min_overlap >= patch_size[1]:
        raise ValueError("!!")
    hs = list(range(0, image_shape[0], patch_size[0] - min_overlap))
    ws = list(range(0, image_shape[1], patch_size[1] - min_overlap))
    # Make sure the final patch is flush with the image boundary
    hs[-1] = image_shape[0] - patch_size[0]
    ws[-1] = image_shape[1] - patch_size[1]
    return [(h, w) for h in hs for w in ws]


def compute_weight(
    hws,
    image_shape,
    patch_size,
    sigma=1.0,
    wtype="gaussian",
    device: torch.device = torch.device("cpu"),
):
    patch_num = len(hws)
    h, w = torch.meshgrid(
        torch.arange(patch_size[0], device=device),
        torch.arange(patch_size[1], device=device),
        indexing="ij",
    )
    h, w = h / float(patch_size[0]), w / float(patch_size[1])
    c_h, c_w = 0.5, 0.5
    h, w = h - c_h, w - c_w
    weights_hw = (h**2 + w**2) ** 0.5 / sigma
    denorm = 1 / (sigma * math.sqrt(2 * math.pi))
    weights_hw = denorm * torch.exp(-0.5 * (weights_hw) ** 2)

    weights = torch.zeros(1, patch_num, *image_shape, device=device)
    for idx, (h, w) in enumerate(hws):
        weights[:, idx, h : h + patch_size[0], w : w + patch_size[1]] = weights_hw
    patch_weights = []
    for idx, (h, w) in enumerate(hws):
        patch_weights.append(
            weights[:, idx : idx + 1, h : h + patch_size[0], w : w + patch_size[1]]
        )

    return patch_weights



import math

import torch
import torch.nn as nn


def positionalencoding2d(pos_embed_dim, height, width):
    """
    :param pos_embed_dim: dimension of the model embeddings
    :param height: height of the positions
    :param width: width of the positions
    :return: height * width * pos_embed_dim matrix
    """
    if pos_embed_dim % 4 != 0:
        raise ValueError(
            "Cannot use sin/cos positional encoding with "
            "odd dimension (got dim={:d})".format(pos_embed_dim)
        )
    pe = torch.zeros(pos_embed_dim, height, width)
    # Each dimension use half of pos_embed_dim
    pos_embed_dim = int(pos_embed_dim / 2)
    div_term = torch.exp(
        torch.arange(0.0, pos_embed_dim, 2) * -(math.log(10000.0) / pos_embed_dim)
    )
    pos_w = torch.arange(0.0, width).unsqueeze(1)
    pos_h = torch.arange(0.0, height).unsqueeze(1)
    pe[0:pos_embed_dim:2, :, :] = (
        torch.sin(pos_w * div_term).transpose(0, 1).unsqueeze(1).repeat(1, height, 1)
    )
    pe[1:pos_embed_dim:2, :, :] = (
        torch.cos(pos_w * div_term).transpose(0, 1).unsqueeze(1).repeat(1, height, 1)
    )
    pe[pos_embed_dim::2, :, :] = (
        torch.sin(pos_h * div_term).transpose(0, 1).unsqueeze(2).repeat(1, 1, width)
    )
    pe[pos_embed_dim + 1 :: 2, :, :] = (
        torch.cos(pos_h * div_term).transpose(0, 1).unsqueeze(2).repeat(1, 1, width)
    )
    pe = pe.permute(1, 2, 0)
    return pe


class RandPosEmbedder(nn.Module):
    def __init__(self, pos_dim, pos_embed_dim, shape, affine):
        super().__init__()
        self.pos_dim = pos_dim
        self.pos_embed_dim = pos_embed_dim
        height, width = shape
        self.pos_embed = nn.Embedding(height * width, pos_embed_dim)
        self.pos_embed_norm_layer = nn.LayerNorm(
            self.pos_embed_dim, eps=1e-12, elementwise_affine=affine
        )
        print("Random discrete embedder for positional encoding ablation created")

    def forward(self, pos_normed):
        B, N, D = pos_normed.shape
        pos_embed_1 = self.pos_embed.weight
        pos_embed_out_1 = self.pos_embed_norm_layer(pos_embed_1)
        pos_embed_out = pos_embed_out_1.unsqueeze(0).repeat((B, 1, 1))
        return pos_embed_out


class SinuPosEmbedder(nn.Module):
    def __init__(self, pos_dim, pos_embed_dim, shape, affine):
        super().__init__()
        self.pos_dim = pos_dim
        self.pos_embed_dim = pos_embed_dim
        self.pos_embed = positionalencoding2d(pos_embed_dim, shape[0], shape[1])
        self.pos_embed = self.pos_embed.cuda().reshape(
            (shape[0] * shape[1], pos_embed_dim)
        )
        print("Sinu embedder for positional encoding ablation created")

    def forward(self, pos_normed):
        B, N, D = pos_normed.shape
        pos_embed_out = self.pos_embed.unsqueeze(0).repeat((B, 1, 1))
        return pos_embed_out


class ZeroEmbedder(nn.Module):
    def __init__(self, pos_embed_dim):
        super().__init__()
        self.pos_embed_dim = pos_embed_dim
        print("Zero embedder for positional encoding ablation created")

    def forward(self, pos_normed):
        B, N, D = pos_normed.shape
        zero_pos_embed = torch.zeros(
            B, N, self.pos_embed_dim, requires_grad=False
        ).cuda()
        return zero_pos_embed


# MM*Mid, MM*Output are the same as in segtran. Just for use by ExpandedFeatTrans.
class MMPrivateMid(nn.Module):
    def __init__(self, config):
        super(MMPrivateMid, self).__init__()
        # Use 1x1 convolution as a group linear layer.
        # Equivalent to each group going through a respective nn.Linear().
        self.num_modes = config.num_modes
        self.feat_dim = config.feat_dim
        feat_dim_allmode = self.feat_dim * self.num_modes
        self.group_linear = nn.Conv1d(
            feat_dim_allmode, feat_dim_allmode, 1, groups=self.num_modes
        )
        self.mid_act_fn = config.act_fun

    def forward(self, x):
        x_trans = self.group_linear(x)  # [B0, 1024*8, 50] -> [B0, 1024*8, 50]
        x_act = self.mid_act_fn(x_trans)  # [B0, 1024*8, 50]
        return x


class MMSharedMid(nn.Module):
    def __init__(self, config):
        super(MMSharedMid, self).__init__()
        self.num_modes = config.num_modes
        self.feat_dim = config.feat_dim
        feat_dim_allmode = self.feat_dim * self.num_modes
        self.shared_linear = nn.Linear(self.feat_dim, self.feat_dim)
        self.mid_act_fn = config.act_fun

    # x: [B0, 1024*8, 50] or [B0, 8, 50, 1024]
    def forward(self, x):
        if len(x.shape) == 3:
            # shape_4d: [B0, 8, 1024, 50].
            shape_4d = (x.shape[0], self.num_modes, self.feat_dim, x.shape[2])
            # x_4d: [B0, 8, 50, 1024].
            x_4d = x.view(shape_4d).permute([0, 1, 3, 2])
            reshaped = True
        else:
            x_4d = x
            reshaped = False

        x_trans = self.shared_linear(x_4d)
        x_act = self.mid_act_fn(x_trans)

        if reshaped:
            # restore the original shape
            x_act = x_act.permute([0, 1, 3, 2]).reshape(x.shape)

        return x_act


# MMPrivateOutput/MMSharedOutput <- ExpandedFeatTrans <- SelfAttFeatTrans <- SegtranFusionEncoder.
# MM***Output has a shortcut (residual) connection.
class MMPrivateOutput(nn.Module):
    def __init__(self, config):
        super(MMPrivateOutput, self).__init__()
        self.num_modes = config.num_modes
        self.feat_dim = config.feat_dim
        feat_dim_allmode = self.feat_dim * self.num_modes
        self.group_linear = nn.Conv1d(
            feat_dim_allmode, feat_dim_allmode, 1, groups=self.num_modes
        )
        self.resout_norm_layer = nn.LayerNorm(
            self.feat_dim, eps=1e-12, elementwise_affine=True
        )
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    # x, shortcut: [B0, 1024*8, 50]
    def forward(self, x, shortcut):
        x = self.group_linear(x)
        # x_comb: [B0, 1024*8, 50]. Residual connection.
        x_comb = x + shortcut
        shape_4d = (x.shape[0], self.num_modes, self.feat_dim, x.shape[2])
        # x_comb_4d, x_drop_4d: [B0, 8, 50, 1024].
        x_comb_4d = x.view(shape_4d).permute([0, 1, 3, 2])
        x_drop_4d = self.dropout(x_comb_4d)
        x_normed = self.resout_norm_layer(x_drop_4d)
        return x_normed


# MMPrivateOutput/MMSharedOutput <- ExpandedFeatTrans <- SelfAttFeatTrans <- SegtranFusionEncoder.
# MM***Output has a shortcut (residual) connection.
class MMSharedOutput(nn.Module):
    # feat_dim_allmode is not used. Just to keep the ctor arguments the same as MMPrivateOutput.
    def __init__(self, config):
        super(MMSharedOutput, self).__init__()
        self.num_modes = config.num_modes
        self.feat_dim = config.feat_dim
        self.shared_linear = nn.Linear(self.feat_dim, self.feat_dim)
        self.resout_norm_layer = nn.LayerNorm(
            self.feat_dim, eps=1e-12, elementwise_affine=True
        )
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    # x, shortcut: [B0, 1024*8, 50] or [B0, 8, 50, 1024]
    def forward(self, x, shortcut):
        # shape_4d: [B0, 8, 1024, 50].
        shape_4d = (x.shape[0], self.num_modes, self.feat_dim, x.shape[2])
        if len(x.shape) == 3:
            x_4d = x.view(shape_4d).permute([0, 1, 3, 2])
        else:
            x_4d = x
        if len(shortcut.shape) == 3:
            shortcut_4d = shortcut.view(shape_4d).permute([0, 1, 3, 2])
        else:
            shortcut_4d = shortcut

        # x_4d, shortcut_4d: [B0, 8, 50, 1024].
        x_trans = self.shared_linear(x_4d)
        # x_4d, x_comb: [B0, 8, 50, 1024]. Residual connection.
        x_comb = x_trans + shortcut_4d
        x_drop = self.dropout(x_comb)
        x_normed = self.resout_norm_layer(x_drop)
        return x_normed


# MultiHeadFeatTrans <- SelfAttFeatTrans.
# MultiHeadFeatTrans has a residual connection.
# We "misuse" num_modes for num_heads, to avoid introducing extra config parameters.
class MultiHeadFeatTrans(nn.Module):
    def __init__(self, config, name):
        super(MultiHeadFeatTrans, self).__init__()
        self.config = config
        self.name = name
        self.in_feat_dim = config.in_feat_dim
        self.feat_dim = config.feat_dim
        self.num_modes = config.num_modes
        self.feat_dim_onehead = self.feat_dim // self.num_modes
        self.feat_dim_allhead = self.feat_dim_onehead * self.num_modes
        # first_linear is the value projection in other transformer implementations.
        # The output of first_linear will be divided into num_modes groups.
        # first_linear is always 'private' for each group, i.e.,
        # parameters are not shared (parameter sharing makes no sense).
        self.first_linear = nn.Linear(self.in_feat_dim, self.feat_dim_allhead)

        print(
            "%s: pool_modes_feat=concat, trans_output_type=%s"
            % (self.name, config.trans_output_type)
        )

        # Disable multiple modes for intermediate and output layers.
        config.num_modes = 1
        self.intermediate = MMSharedMid(self.config)

        if config.trans_output_type == "shared":
            self.output = MMSharedOutput(config)
        elif config.trans_output_type == "private":
            self.output = MMPrivateOutput(config)

        self.apply_attn_early = config.apply_attn_early

    def add_identity_bias(self):
        if self.config.feattrans_lin1_idbias_scale > 0:
            identity_weight = (
                torch.diag(torch.ones(self.feat_dim))
                * self.config.initializer_range
                * self.config.feattrans_lin1_idbias_scale
            )
            # Only bias the weight of the first mode.
            # The total initial "weight mass" in each row is reduced by 1024*0.02*0.5.
            self.first_linear.weight.data[: self.feat_dim] = (
                self.first_linear.weight.data[: self.feat_dim] * 0.5 + identity_weight
            )

    def forward(self, input_feat, attention_probs, attention_scores):
        # input_feat: [B0, 50, 1024], mm_first_feat: [B0, 50, 1024*8]
        mm_first_feat = self.first_linear(input_feat)
        # mm_first_feat_act after permute: [B0, 1024*8, 50]
        mm_first_feat = mm_first_feat.permute(0, 2, 1)

        if self.apply_attn_early:
            # shape_4d: [B0, 8, 1024, 50]
            shape_4d = (
                mm_first_feat.shape[0],
                self.num_modes,
                self.feat_dim_onehead,
                mm_first_feat.shape[2],
            )
            # mm_first_feat_4d: [B0, 8, 50, 1024]
            mm_first_feat_4d = mm_first_feat.view(shape_4d).permute([0, 1, 3, 2])
            mm_first_feat_fusion = torch.matmul(attention_probs, mm_first_feat_4d)
            mm_first_feat_fusion_3d = mm_first_feat_fusion.permute(
                [0, 1, 3, 2]
            ).reshape(mm_first_feat.shape)
            mm_first_feat = mm_first_feat_fusion_3d

        # mm_mid_feat:   [B0, 1024*8, 50]. Group linear & gelu of mm_first_feat.
        mm_mid_feat = self.intermediate(mm_first_feat)
        # mm_last_feat:  [B0, 8, 50, 1024]. Group/shared linear & residual & Layernorm
        mm_last_feat = self.output(mm_mid_feat, mm_first_feat)

        if (attention_probs is not None) and (not self.apply_attn_early):
            # matmul(t1, t2): (h1, w1), (w1, w2) => (h1, w2)
            # [B0, 8, 50, 50][B0, 8, 50, 1024] -> mm_trans_feat: [B0, 8, 50, 1024]
            mm_trans_feat = torch.matmul(attention_probs, mm_last_feat)
        else:
            mm_trans_feat = mm_last_feat

        trans_feat = mm_trans_feat.squeeze(1)

        # trans_feat: [B0, 50, 1024],   if pool_modes_feat != 'none',
        # or          [B0, 8, 50, 1024] if pool_modes_feat == 'none'.
        return trans_feat



# =============================================================================
# Copyright 2023 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Code adapted from RAFT: https://github.com/princeton-vl/RAFT/blob/master/core/update.py
# =============================================================================

from functools import partial

import torch
import torch.nn as nn

from .pkconv import PKConv2d
from .pkconv_slk import PKConvSLK
from .local_timm.norm import LayerNorm2d


class FlowHeadPartial(nn.Module):
    def __init__(self, input_dim=128, hidden_dim=256, cache_pkconv_weights=False):
        super(FlowHeadPartial, self).__init__()
        self.conv1 = PKConv2d(
            input_dim, hidden_dim, 3, padding=1, cache_weights=cache_pkconv_weights
        )
        self.conv2 = PKConv2d(
            hidden_dim, 2, 3, padding=1, cache_weights=cache_pkconv_weights
        )
        self.act = nn.ReLU(inplace=True)

    def forward(self, x, out_ch=None):
        return self.conv2(self.act(self.conv1(x)), out_ch)


class ConvPartialGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192 + 128, cache_pkconv_weights=False):
        super(ConvPartialGRU, self).__init__()
        self.convz = PKConv2d(
            hidden_dim + input_dim,
            hidden_dim,
            3,
            padding=1,
            cache_weights=cache_pkconv_weights,
        )
        self.convr = PKConv2d(
            hidden_dim + input_dim,
            hidden_dim,
            3,
            padding=1,
            cache_weights=cache_pkconv_weights,
        )
        self.convq = PKConv2d(
            hidden_dim + input_dim,
            hidden_dim,
            3,
            padding=1,
            cache_weights=cache_pkconv_weights,
        )

    def forward(self, h, x, out_ch):
        hx = torch.cat([h, x], dim=1)

        z = self.convz(hx, out_ch)
        z = torch.sigmoid(z)

        r = self.convr(hx, out_ch)
        r = torch.sigmoid(r)

        q = self.convq(torch.cat([r * h, x], dim=1), out_ch)
        q = torch.tanh(q)

        h = (1 - z) * h + z * q
        return h


class PKConvSLKGRU(nn.Module):
    def __init__(
        self,
        hidden_dim=128,
        input_dim=192 + 128,
        use_norm_affine=False,
        num_groups=8,
        depth=2,
        mlp_ratio=4,
        cache_pkconv_weights=False,
    ):
        super(PKConvSLKGRU, self).__init__()
        norm_layer = partial(LayerNorm2d, affine=use_norm_affine)

        self.convz = PKConvSLK(
            hidden_dim + input_dim,
            hidden_dim,
            mlp_ratio=mlp_ratio,
            norm_layer=norm_layer,
            stride=1,
            depth=depth,
            cache_pkconv_weights=cache_pkconv_weights,
        )
        self.convr = PKConvSLK(
            hidden_dim + input_dim,
            hidden_dim,
            mlp_ratio=mlp_ratio,
            norm_layer=norm_layer,
            stride=1,
            depth=depth,
            cache_pkconv_weights=cache_pkconv_weights,
        )
        self.convq = PKConvSLK(
            hidden_dim + input_dim,
            hidden_dim,
            mlp_ratio=mlp_ratio,
            norm_layer=norm_layer,
            stride=1,
            depth=depth,
            cache_pkconv_weights=cache_pkconv_weights,
        )

    def forward(self, h, x, out_ch):
        hx = torch.cat([h, x], dim=1)

        z = self.convz(hx, out_ch)
        z = torch.sigmoid(z)

        r = self.convr(hx, out_ch)
        r = torch.sigmoid(r)

        q = self.convq(torch.cat([r * h, x], dim=1), out_ch)
        q = torch.tanh(q)

        h = (1 - z) * h + z * q
        return h


class ConvexMask(nn.Module):
    def __init__(self, net_chs, pred_stride, cache_pkconv_weights=False):
        super(ConvexMask, self).__init__()
        self.conv1 = PKConv2d(
            net_chs, net_chs * 2, 3, padding=1, cache_weights=cache_pkconv_weights
        )
        self.act = nn.ReLU(inplace=True)
        self.conv2 = PKConv2d(
            net_chs * 2,
            pred_stride**2 * 9,
            1,
            padding=0,
            cache_weights=cache_pkconv_weights,
        )

    def forward(self, x):
        in_chs = x.shape[1]
        x = self.conv1(x, 2 * in_chs)
        x = self.act(x)
        x = self.conv2(x)
        return x


class MotionEncoderPartial(nn.Module):
    def __init__(
        self,
        corr_levels: int,
        corr_range: int,
        dec_motion_chs: int,
        cache_pkconv_weights: bool,
    ):
        super(MotionEncoderPartial, self).__init__()

        c_hidden = 256
        c_out = 192
        f_hidden = 128
        f_out = 64

        cor_planes = corr_levels * (2 * corr_range + 1) ** 2
        self.convc1 = PKConv2d(
            cor_planes, c_hidden, 1, padding=0, cache_weights=cache_pkconv_weights
        )
        self.convc2 = PKConv2d(
            c_hidden, c_out, 3, padding=1, cache_weights=cache_pkconv_weights
        )

        self.convf1 = PKConv2d(
            2, f_hidden, 7, padding=3, cache_weights=cache_pkconv_weights
        )
        self.convf2 = PKConv2d(
            f_hidden, f_out, 3, padding=1, cache_weights=cache_pkconv_weights
        )

        in_ch = f_out + c_out
        out_ch = dec_motion_chs - 2
        self.conv = PKConv2d(
            in_ch, out_ch, 3, padding=1, cache_weights=cache_pkconv_weights
        )

        self.act = nn.ReLU()

    def forward(self, flow, corr):
        cor = self.act(self.convc1(corr))
        cor = self.act(self.convc2(cor))

        outs = [cor]

        flo = self.act(self.convf1(flow))
        flo = self.act(self.convf2(flo))
        outs.append(flo)

        outs = torch.cat(outs, dim=1)
        out = self.act(self.conv(outs))
        out_t = [out, flow]
        return torch.cat(out_t, dim=1)


class UpdatePartialBlock(nn.Module):
    def __init__(
        self,
        pyramid_ranges: tuple[int, int],
        corr_levels: int,
        corr_range: int,
        net_chs_fixed: int,
        inp_chs_fixed: int,
        group_norm_num_groups: int,
        use_norm_affine: bool,
        dec_motion_chs: int,
        dec_gru_depth: int,
        dec_gru_iters: int,
        dec_gru_mlp_ratio: float,
        use_upsample_mask: bool,
        upmask_gradient_scale: float,
        cache_pkconv_weights: bool,
    ):
        super(UpdatePartialBlock, self).__init__()

        self.use_upsample_mask = use_upsample_mask
        self.upmask_gradient_scale = upmask_gradient_scale

        self.encoder = MotionEncoderPartial(
            corr_levels=corr_levels,
            corr_range=corr_range,
            dec_motion_chs=dec_motion_chs,
            cache_pkconv_weights=cache_pkconv_weights,
        )
        self.gru_list = nn.ModuleList(
            [
                PKConvSLKGRU(
                    hidden_dim=net_chs_fixed,
                    input_dim=dec_motion_chs + inp_chs_fixed,
                    use_norm_affine=use_norm_affine,
                    num_groups=group_norm_num_groups,
                    depth=dec_gru_depth,
                    mlp_ratio=dec_gru_mlp_ratio,
                    cache_pkconv_weights=cache_pkconv_weights,
                )
                for _ in range(dec_gru_iters)
            ]
        )

        self.flow_head = FlowHeadPartial(
            net_chs_fixed,
            hidden_dim=256,
            cache_pkconv_weights=cache_pkconv_weights,
        )

        if self.use_upsample_mask:
            pred_stride = min(pyramid_ranges)
            self.mask = ConvexMask(
                net_chs_fixed,
                pred_stride,
                cache_pkconv_weights=cache_pkconv_weights,
            )

    def forward(self, net, inp, corr, flow):
        motion_features = self.encoder(flow, corr)

        inp = torch.cat([inp, motion_features], dim=1)

        for gru in self.gru_list:
            net = gru(net, inp, net.shape[1])

        delta_flow = self.flow_head(net)

        mask = None
        if self.use_upsample_mask:
            mask = self.upmask_gradient_scale * self.mask(net)

        return delta_flow, net, mask



import torch
import torch.nn as nn

from .attention import (
    single_head_full_attention,
    single_head_split_window_attention,
    single_head_full_attention_1d,
    single_head_split_window_attention_1d,
)
from .utils import generate_shift_window_attn_mask, generate_shift_window_attn_mask_1d


class TransformerLayer(nn.Module):
    def __init__(
        self,
        d_model=128,
        nhead=1,
        no_ffn=False,
        ffn_dim_expansion=4,
    ):
        super(TransformerLayer, self).__init__()

        self.dim = d_model
        self.nhead = nhead
        self.no_ffn = no_ffn

        # multi-head attention
        self.q_proj = nn.Linear(d_model, d_model, bias=False)
        self.k_proj = nn.Linear(d_model, d_model, bias=False)
        self.v_proj = nn.Linear(d_model, d_model, bias=False)

        self.merge = nn.Linear(d_model, d_model, bias=False)

        self.norm1 = nn.LayerNorm(d_model)

        # no ffn after self-attn, with ffn after cross-attn
        if not self.no_ffn:
            in_channels = d_model * 2
            self.mlp = nn.Sequential(
                nn.Linear(in_channels, in_channels * ffn_dim_expansion, bias=False),
                nn.GELU(),
                nn.Linear(in_channels * ffn_dim_expansion, d_model, bias=False),
            )

            self.norm2 = nn.LayerNorm(d_model)

    def forward(
        self,
        source,
        target,
        height=None,
        width=None,
        shifted_window_attn_mask=None,
        shifted_window_attn_mask_1d=None,
        attn_type="swin",
        with_shift=False,
        attn_num_splits=None,
    ):
        # source, target: [B, L, C]
        query, key, value = source, target, target

        # for stereo: 2d attn in self-attn, 1d attn in cross-attn
        is_self_attn = (query - key).abs().max() < 1e-6

        # single-head attention
        query = self.q_proj(query)  # [B, L, C]
        key = self.k_proj(key)  # [B, L, C]
        value = self.v_proj(value)  # [B, L, C]

        if (
            attn_type == "swin" and attn_num_splits > 1
        ):  # self, cross-attn: both swin 2d
            if self.nhead > 1:
                # we observe that multihead attention slows down the speed and increases the memory consumption
                # without bringing obvious performance gains and thus the implementation is removed
                raise NotImplementedError
            else:
                message = single_head_split_window_attention(
                    query,
                    key,
                    value,
                    num_splits=attn_num_splits,
                    with_shift=with_shift,
                    h=height,
                    w=width,
                    attn_mask=shifted_window_attn_mask,
                )

        elif (
            attn_type == "self_swin2d_cross_1d"
        ):  # self-attn: swin 2d, cross-attn: full 1d
            if self.nhead > 1:
                raise NotImplementedError
            else:
                if is_self_attn:
                    if attn_num_splits > 1:
                        message = single_head_split_window_attention(
                            query,
                            key,
                            value,
                            num_splits=attn_num_splits,
                            with_shift=with_shift,
                            h=height,
                            w=width,
                            attn_mask=shifted_window_attn_mask,
                        )
                    else:
                        # full 2d attn
                        message = single_head_full_attention(
                            query, key, value
                        )  # [N, L, C]

                else:
                    # cross attn 1d
                    message = single_head_full_attention_1d(
                        query,
                        key,
                        value,
                        h=height,
                        w=width,
                    )

        elif (
            attn_type == "self_swin2d_cross_swin1d"
        ):  # self-attn: swin 2d, cross-attn: swin 1d
            if self.nhead > 1:
                raise NotImplementedError
            else:
                if is_self_attn:
                    if attn_num_splits > 1:
                        # self attn shift window
                        message = single_head_split_window_attention(
                            query,
                            key,
                            value,
                            num_splits=attn_num_splits,
                            with_shift=with_shift,
                            h=height,
                            w=width,
                            attn_mask=shifted_window_attn_mask,
                        )
                    else:
                        # full 2d attn
                        message = single_head_full_attention(
                            query, key, value
                        )  # [N, L, C]
                else:
                    if attn_num_splits > 1:
                        assert shifted_window_attn_mask_1d is not None
                        # cross attn 1d shift
                        message = single_head_split_window_attention_1d(
                            query,
                            key,
                            value,
                            num_splits=attn_num_splits,
                            with_shift=with_shift,
                            h=height,
                            w=width,
                            attn_mask=shifted_window_attn_mask_1d,
                        )
                    else:
                        message = single_head_full_attention_1d(
                            query,
                            key,
                            value,
                            h=height,
                            w=width,
                        )

        else:
            message = single_head_full_attention(query, key, value)  # [B, L, C]

        message = self.merge(message)  # [B, L, C]
        message = self.norm1(message)

        if not self.no_ffn:
            message = self.mlp(torch.cat([source, message], dim=-1))
            message = self.norm2(message)

        return source + message


class TransformerBlock(nn.Module):
    """self attention + cross attention + FFN"""

    def __init__(
        self,
        d_model=128,
        nhead=1,
        ffn_dim_expansion=4,
    ):
        super(TransformerBlock, self).__init__()

        self.self_attn = TransformerLayer(
            d_model=d_model,
            nhead=nhead,
            no_ffn=True,
            ffn_dim_expansion=ffn_dim_expansion,
        )

        self.cross_attn_ffn = TransformerLayer(
            d_model=d_model,
            nhead=nhead,
            ffn_dim_expansion=ffn_dim_expansion,
        )

    def forward(
        self,
        source,
        target,
        height=None,
        width=None,
        shifted_window_attn_mask=None,
        shifted_window_attn_mask_1d=None,
        attn_type="swin",
        with_shift=False,
        attn_num_splits=None,
    ):
        # source, target: [B, L, C]

        # self attention
        source = self.self_attn(
            source,
            source,
            height=height,
            width=width,
            shifted_window_attn_mask=shifted_window_attn_mask,
            attn_type=attn_type,
            with_shift=with_shift,
            attn_num_splits=attn_num_splits,
        )

        # cross attention and ffn
        source = self.cross_attn_ffn(
            source,
            target,
            height=height,
            width=width,
            shifted_window_attn_mask=shifted_window_attn_mask,
            shifted_window_attn_mask_1d=shifted_window_attn_mask_1d,
            attn_type=attn_type,
            with_shift=with_shift,
            attn_num_splits=attn_num_splits,
        )

        return source


class FeatureTransformer(nn.Module):
    def __init__(
        self,
        num_layers=6,
        d_model=128,
        nhead=1,
        ffn_dim_expansion=4,
    ):
        super(FeatureTransformer, self).__init__()

        self.d_model = d_model
        self.nhead = nhead

        self.layers = nn.ModuleList(
            [
                TransformerBlock(
                    d_model=d_model,
                    nhead=nhead,
                    ffn_dim_expansion=ffn_dim_expansion,
                )
                for i in range(num_layers)
            ]
        )

        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def forward(
        self,
        feature0,
        feature1,
        attn_type="swin",
        attn_num_splits=None,
        **kwargs,
    ):
        b, c, h, w = feature0.shape
        assert self.d_model == c

        feature0 = feature0.flatten(-2).permute(0, 2, 1)  # [B, H*W, C]
        feature1 = feature1.flatten(-2).permute(0, 2, 1)  # [B, H*W, C]

        # 2d attention
        if "swin" in attn_type and attn_num_splits > 1:
            # global and refine use different number of splits
            window_size_h = h // attn_num_splits
            window_size_w = w // attn_num_splits

            # compute attn mask once
            shifted_window_attn_mask = generate_shift_window_attn_mask(
                input_resolution=(h, w),
                window_size_h=window_size_h,
                window_size_w=window_size_w,
                shift_size_h=window_size_h // 2,
                shift_size_w=window_size_w // 2,
                device=feature0.device,
            )  # [K*K, H/K*W/K, H/K*W/K]
        else:
            shifted_window_attn_mask = None

        # 1d attention
        if "swin1d" in attn_type and attn_num_splits > 1:
            window_size_w = w // attn_num_splits

            # compute attn mask once
            shifted_window_attn_mask_1d = generate_shift_window_attn_mask_1d(
                input_w=w,
                window_size_w=window_size_w,
                shift_size_w=window_size_w // 2,
                device=feature0.device,
            )  # [K, W/K, W/K]
        else:
            shifted_window_attn_mask_1d = None

        # concat feature0 and feature1 in batch dimension to compute in parallel
        concat0 = torch.cat((feature0, feature1), dim=0)  # [2B, H*W, C]
        concat1 = torch.cat((feature1, feature0), dim=0)  # [2B, H*W, C]

        for i, layer in enumerate(self.layers):
            concat0 = layer(
                concat0,
                concat1,
                height=h,
                width=w,
                attn_type=attn_type,
                with_shift="swin" in attn_type and attn_num_splits > 1 and i % 2 == 1,
                attn_num_splits=attn_num_splits,
                shifted_window_attn_mask=shifted_window_attn_mask,
                shifted_window_attn_mask_1d=shifted_window_attn_mask_1d,
            )

            # update feature1
            concat1 = torch.cat(concat0.chunk(chunks=2, dim=0)[::-1], dim=0)

        feature0, feature1 = concat0.chunk(chunks=2, dim=0)  # [B, H*W, C]

        # reshape back
        feature0 = (
            feature0.view(b, h, w, c).permute(0, 3, 1, 2).contiguous()
        )  # [B, C, H, W]
        feature1 = (
            feature1.view(b, h, w, c).permute(0, 3, 1, 2).contiguous()
        )  # [B, C, H, W]

        return feature0, feature1



"""Basic functions to handle optical flow."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import json
from pathlib import Path
from typing import Any, IO, Optional, Sequence, Union

import cv2 as cv
import numpy as np
import torch

from .external import flowpy, raft, selflow, flow_IO
from . import flowpy_torch


def flow_to_rgb(
    flow: Union[np.ndarray, torch.Tensor],
    flow_max_radius: Optional[Union[float, torch.Tensor]] = None,
    background: str = "bright",
    custom_colorwheel: Optional[torch.Tensor] = None,
) -> Union[np.ndarray, torch.Tensor]:
    """Convert flows to RGB images.

    The input can be either numpy or torch tensors. This function is just a wrapper for flowpy and flowpy_torch.

    Parameters
    ----------
    flow : np.ndarray or torch.Tensor
        If flow is a numpy array, then it must have 3 dimensions HWC (Height, Width, Channels) - notice it is channels last.
        If it is a torch tensor, then it has at least 3 dimensions in the ...CHW (..., Channels, Height, Width) layout,
        where ... represents any number of dimensions.
        Channel 0 should be the x-displacement.
        Channel 1 should be the y-displacement.
    flow_max_radius : float or torch.Tensor, optional
        Set the radius that gives the maximum color intensity, useful for comparing different flows.
        Default: The normalization is based on the input flow maximum radius per batch element.
    background : str, default 'bright'
        States if zero-valued flow should look 'bright' or 'dark'.
    custom_colorwheel : np.ndarray or torch.Tensor
        Use a custom colorwheel for specific hue transition lengths. By default, the default transition lengths are used.

    Returns
    -------
    np.ndarray or torch.Tensor
        The RGB image representing the flow. It keeps the same dimensions and type as the input.

    See Also
    --------
    ptlflow.utils.external.flowpy.flow_to_rgb
    ptlflow.utils.flowpy_torch.flow_to_rgb
    """
    if isinstance(flow, np.ndarray):
        flow_rgb = flowpy.flow_to_rgb(
            flow, flow_max_radius, background, custom_colorwheel
        )
    else:
        flow_rgb = flowpy_torch.flow_to_rgb(
            flow, flow_max_radius, background, custom_colorwheel
        )
    return flow_rgb


def flow_read(
    input_data: Union[Sequence[Any], str, Path, IO],
    format: Optional[str] = None,
) -> np.ndarray:
    """Read optical flow from file.

    This is just a wrapper for flowpy (for .flo and .png) or raft (for pfm), added for convenience.

    Parameters
    ----------
    input_data: Sequence[Any], str, Path or IO
        Path of the file to read or a sequence containing the path and extra information.
    format: str, optional
        Specify in what format the flow is read, accepted formats: "flo", "flo5", "kubric_png", "npz", "pfm", "png".
        If None, it is guessed from the file extension.

    Returns
    -------
    numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] is the x-displacement.
        flow[..., 1] is the y-displacement.

    See Also
    --------
    ptlflow.utils.external.flowpy.flow_read
    ptlflow.utils.external.raft.read_pfm
    ptlflow.utils.external.flow_IO.readFlo5Flow
    write_pfm
    """
    if (format is not None and format == "pfm") or str(input_data).endswith("pfm"):
        return raft.read_pfm(input_data)
    elif (format is not None and format == "flo5") or str(input_data).endswith("flo5"):
        return flow_IO.readFlo5Flow(input_data)
    elif (format is not None and format == "npy") or str(input_data).endswith("npy"):
        return np.load(input_data)
    elif format is not None and format == "kubric_png":
        return read_kubric_flow(input_data[0], input_data[1])
    elif format is not None and format == "viper_npz":
        return read_viper_flow(input_data)
    else:
        return flowpy.flow_read(input_data, format)


def flow_write(
    output_file: Union[str, Path, IO], flow: np.ndarray, format: str = None
) -> None:
    """Write optical flow to file.

    This is just a wrapper for flowpy (for .flo and .png) or selflow (for pfm), added for convenience.

    Parameters
    ----------
    output_file: str, Path or IO
        Path of the file to write or file object.
    flow: numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] should be the x-displacement
        flow[..., 1] should be the y-displacement
    format: str, optional
        Specify in what format the flow is written, accepted formats: "png" or "flo"
        If None, it is guessed on the file extension

    See Also
    --------
    ptlflow.utils.external.flowpy.flow_write
    """
    if (format is not None and format == "pfm") or str(output_file).endswith("pfm"):
        selflow.write_pfm(output_file, flow)
    elif (format is not None and format == "flo5") or str(output_file).endswith("flo5"):
        flow_IO.writeFlo5File(flow, output_file)
    elif (format is not None and format == "npy") or str(output_file).endswith("npy"):
        np.save(output_file, flow)
    elif format is not None and format == "viper_npz":
        return write_viper_flow(output_file, flow)
    else:
        flowpy.flow_write(output_file, flow, format)


def read_kubric_flow(
    input_file: Union[str, Path, IO],
    flow_direction: str,
) -> np.ndarray:
    """Read optical flow in Kubric PNG format from file.

    Parameters
    ----------
    input_file: str, Path or IO
        Path of the file to read or file object.
    flow_direction: str
        Either "backward_flow" or "forward_flow".

    Returns
    -------
    numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] is the x-displacement.
        flow[..., 1] is the y-displacement.
    """
    with open(Path(input_file).parent / "data_ranges.json", "r") as f:
        data_ranges = json.load(f)
    flow_min = data_ranges[flow_direction]["min"]
    flow_max = data_ranges[flow_direction]["max"]

    flow = cv.imread(str(input_file), cv.IMREAD_UNCHANGED)[..., 1:].astype(np.float32)
    flow = flow / 65535 * (flow_max - flow_min) + flow_min
    return flow


def read_viper_flow(input_file: Union[str, Path, IO]) -> np.ndarray:
    """Read optical flow in VIPER npz format from file.

    Parameters
    ----------
    input_file: str, Path or IO
        Path of the file to read or file object.

    Returns
    -------
    numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] is the x-displacement.
        flow[..., 1] is the y-displacement.
    """
    flow_npz = np.load(input_file)
    flow = np.stack([flow_npz["u"], flow_npz["v"]], 2).astype(np.float32)
    flow[np.abs(flow) > 512] = np.nan
    return flow


def fb_check(
    forward_flow: Union[np.ndarray, torch.Tensor],
    backward_flow: Union[np.ndarray, torch.Tensor],
    threshold: float = 1.0,
):
    is_np_input = False
    if isinstance(forward_flow, np.ndarray):
        assert len(forward_flow.shape) == 3
        forward_flow = torch.from_numpy(forward_flow).permute(2, 0, 1)[None]
        backward_flow = torch.from_numpy(backward_flow).permute(2, 0, 1)[None]
        is_np_input = True
    assert len(forward_flow.shape) == 4

    coords = torch.meshgrid(
        torch.arange(forward_flow.shape[-2], dtype=torch.float32),
        torch.arange(forward_flow.shape[-1], dtype=torch.float32),
        indexing="ij",
    )
    coords = torch.stack(coords[::-1], dim=0)[None]

    coords = coords + forward_flow
    coords = coords.permute(0, 2, 3, 1)
    warped_backward_flow, in_mask = raft.bilinear_sampler(
        backward_flow, coords, mask=True
    )
    fb_diff = torch.norm(forward_flow + warped_backward_flow, p=2, dim=1)
    fb_mask = (fb_diff < threshold) & (in_mask[..., 0] > 0.5)

    if is_np_input:
        fb_mask = fb_mask[0].numpy()
    return fb_mask


def write_viper_flow(output_file: Union[str, Path], flow: np.ndarray):
    flow = flow.astype(np.float16)
    np.savez(output_file, u=flow[..., 0], v=flow[..., 1])
